# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Sim Bamford
         Aiko Dinale

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation
                            and Export of Events etc)
Functions for manipulating 3D 6DOF poses

A set of functions for manipulating dictionaries containing timestamped sets
of 3D poses.
The typical dictionary format is:

{
*	'ts': numpy array (n) of float - timestamps in seconds, assumed to be
                            sorted ascending,
*	'point': numpy array (n, 3) of float, where each row contains [x, y, z],
*	'rotation': see below for different pose formats,
*   optionally, other fields, where each is an iterable with 1st dimension (n, ...)
}

So the nth point and the nth rotation represent the pose at the nth timestamp.

The rotation may be either:

* Unit quaternion: (n, 4) where each row contains [w, x, y, z] (NOTE! ordering
                          of quaternion elements may differ in other software
                          frameworks, e.g. ROS uses [x, y, z, w]
* Rotation matrix: (n, 3, 3)
* Rodrigues vector: (n, 3)
* Axis-Angle: (n, 1) this is the angle of axis-angle representation, there is
                     then a further field called 'axis' (n, 3)

We will name these formats differently depending on the rotation representation,
respectively:

pose6q
pose6mat
pose6rvec
pose6axang

In the case of 'mat', there is a further field 'transformation', (n, 4, 4)
which contains transformation matrices for each timestamp. In fact, the 'point'
and 'rotation' fields in this case are simply views to the appropriate parts
of 'transformation', i.e. [n, :3, 3] and [n, :3, :3] respectively.

There are functions to translate one rotation representation to another:

rotToMat(dict)
rotToVec(dict)
rotToQuat(dict)
rotToAxisAngle(dict)

Each of the above deduces the type from the dimensions of 'rotation' before
applying the appropriate transformation.

There are these additional functions, which apply to all samples in a pose dict:

interpolatePoses(dict, times=None, period=None, maxPeriod=None)
rotate(dict, rotation)
translate(dict, translation)
transform(dict, translation=None, rotation=None)
angleBetween(dict, rotation)
averageRotation(dict, weights=None)

There are also these forms, which work sample-wise between two dicts:

transform(dict1, rotation=dict2) - applies as if the dicts contained mats
                                    and the result were: dict2 @ dict1
angleBetween(dict1, rotation=dict2)

In both of the above cases, if the dicts do not have identical timstamps,
then the resulting dict has a result at each time in each dict
which is contained within the bounds of the timestamps in the other dict
- i.e. interpolation is used to align the data intime.

There are lower level functions which don't operate on dicts but rather
on the appropriate arrays and on singular samples; these are exposed and can
be imported for convenience.
"""

import numpy as np

# local imports
from .timestamps import sortDataTypeDictByTime
from .split import selectByBool

#%% Helper functions

def findRotationRepresentation(rotation):
    if rotation.shape[1] == 1:
        return 'axisAngle'
    elif rotation.shape[1] == 4:
        return 'quaternion'
    elif rotation.ndim == 2:
        return 'rVec'
    else:
        return 'mat'

def findRotationRepresentationSingle(rotation):
    if rotation.shape[0] == 4:
        return 'quaternion'
    elif rotation.ndim == 2:
        return 'mat'
    else:
        return 'rVec'

#%% Functions we want

def rotToRepresentation(inDict, rotationRepresentation):
    if rotationRepresentation == 'quaternion':
        return rotToQuat(inDict)
    elif rotationRepresentation == 'axisAngle':
        return rotToAxisAngle(inDict)
    elif rotationRepresentation == 'rVec':
        return rotToVec(inDict)
    else: #rotationRepresentation == 'mat'
        return rotToMat(inDict)


def rotToMat(inDict):
    rotation = inDict['rotation']
    rotationRepresentation = findRotationRepresentation(rotation)
    if rotationRepresentation == 'mat':
        # already a matrix
        return inDict

    mat = np.zeros((rotation.shape[0], 4, 4), dtype=float)
    mat[:, 3, 3] = 1
    mat[:, :3, 3] = inDict['point']
    outDict = inDict.copy()
    outDict['point'] = mat[:, :3, 3]
    outDict['rotation'] = mat[:, :3, :3]
    outDict['transformation'] = mat

    if rotationRepresentation == 'axisAngle':
        angle = rotation
        axis = inDict['axis']
        x = axis[:, 0]
        y = axis[:, 1]
        z = axis[:, 2]
        sin = np.sin(angle)
        cos = np.cos(angle)
        outDict = inDict.copy()
        # following: https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
        mat[:, 0, 0] = cos + x**2 * (1 - cos)
        mat[:, 0, 1] = x * y * (1 - cos) - z * sin
        mat[:, 0, 2] = x * z * (1 - cos) + y * sin
        mat[:, 1, 0] = y * x * (1 - cos) + z * sin
        mat[:, 1, 1] = cos + y**2 * (1 - cos)
        mat[:, 1, 2] = y * z * (1 - cos) - x * sin
        mat[:, 2, 0] = z * x * (1 - cos) - y * sin
        mat[:, 2, 1] = z * y * (1 - cos) + x * sin
        mat[:, 2, 2] = cos + z**2 * (1 - cos)
    elif rotationRepresentation == 'quaternion':
        w = rotation[:, 0]
        x = rotation[:, 1]
        y = rotation[:, 2]
        z = rotation[:, 3]
        mat[:, 0, 0] = 1 - 2 * y**2 - 2 * z**2
        mat[:, 0, 1] = 2 * x * y - 2 * z * w
        mat[:, 0, 2] = 2 * x * z + 2 * y * w
        mat[:, 1, 0] = 2 * x * y + 2 * z * w
        mat[:, 1, 1] = 1 - 2 * x**2 - 2 * z**2
        mat[:, 1, 2] = 2 * y * z - 2 * x*w
        mat[:, 2, 0] = 2 * x * z - 2 * y*w
        mat[:, 2, 1] = 2 * y * z + 2 * x*w
        mat[:, 2, 2] = 1 - 2 * x**2 - 2 * y**2
    else: #rotationRepresentation == 'rVec'
        # following: https://github.com/robEllenberg/comps-plugins/blob/master/python/rodrigues.py
        # Start with identity matrix
        '''
        mat[:, 0, 0] = 1
        mat[:, 1, 1] = 1
        mat[:, 1, 2] = 1
        def S(n, mat):
            rotation
            Sn = np.zeros_like(mat)
            Sn[:, 0, 1] = -n[:, 2]
            Sn[:, 0, 2] = n[:, 1]
            Sn[:, 1, 0] = n[:, 2]
            Sn[:, 0, 2] = -n[:, 0]
            Sn[:, 2, 0] = -n[:, 1]
            Sn[:, 2, 1] = n[:, 0]
            return Sn

        theta = np.linalg.norm(rotation, axis=1)
        n = rotation / theta
        Sn = S(n, mat)
        Sr = S(r, mat)
        theta2 = theta**2
        cond = theta > 1e-30

        def dotAlongZerothAxis(S, mat): # TODO: there is probably a neat numpy way of doing this: tensordot? apply_along_axis?
            out = np.zeros_like(mat)
            for idx, slicee in enumerate(S):
                out[idx, :, :] = np.dot(S, S)
            return out

        mat = mat \
            + cond * ( np.sin(theta)*Sn + (1-np.cos(theta))*dotAlongZerothAxis(Sn, mat) ) \
            + (1-cond) * ((1-theta2/6)*Sr + (0.5-theta2/24)*dotAlongZerothAxis(Sr, mat) )
        '''
        pass
    return outDict

def rotToVec(inDict):
    rotation = inDict['rotation']
    rotationRepresentation = findRotationRepresentation(rotation)
    if rotationRepresentation == 'rVec':
        return inDict
    outDict = inDict.copy()
    if rotationRepresentation == 'axisAngle':
        raise NotImplementedError("Method not implemented")
    elif rotationRepresentation == 'quaternion':
        raise NotImplementedError("Method not implemented")
    elif rotationRepresentation == 'rVec':
        raise NotImplementedError("Method not implemented")
    else: #rotationRepresentation == 'mat'
        raise NotImplementedError("Method not implemented")
    return outDict

def rotToQuat(inDict):
    rotation = inDict['rotation']
    rotationRepresentation = findRotationRepresentation(rotation)
    if rotationRepresentation == 'quaternion':
        return inDict
    if rotationRepresentation == 'axisAngle':
        raise NotImplementedError("Method not implemented")
    elif rotationRepresentation == 'rVec':
        raise NotImplementedError("Method not implemented")
    else: #rotationRepresentation == 'mat'
        outDict = inDict.copy()
        # Solution from https://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/
        m = rotation
        m00 = m[:, 0, 0]
        m01 = m[:, 0, 1]
        m02 = m[:, 0, 2]
        m10 = m[:, 1, 0]
        m11 = m[:, 1, 1]
        m12 = m[:, 1, 2]
        m20 = m[:, 2, 0]
        m21 = m[:, 2, 1]
        m22 = m[:, 2, 2]
        tr = m00 + m11 + m22
        # TODO: it should be possible to matricise the following
        numRotations = m.shape[0]
        qw = np.zeros(numRotations)
        qx = np.zeros(numRotations)
        qy = np.zeros(numRotations)
        qz = np.zeros(numRotations)
        for i in range(m.shape[0]):
            if tr[i] > 0:
                S = np.sqrt(tr[i] + 1.0) * 2
                qw[i] = 0.25 * S
                qx[i] = (m21[i] - m12[i]) / S
                qy[i] = (m02[i] - m20[i]) / S
                qz[i] = (m10[i] - m01[i]) / S
            elif (m00[i] > m11[i]) and(m00[i] > m22[i]):
                S = np.sqrt(1.0 + m00[i] - m11[i] - m22[i]) * 2
                qw[i] = (m21[i] - m12[i]) / S
                qx[i] = 0.25 * S
                qy[i] = (m01[i] + m10[i]) / S
                qz[i] = (m02[i] + m20[i]) / S
            elif m11[i] > m22[i]:
                S = np.sqrt(1.0 + m11[i] - m00[i] - m22[i]) * 2
                qw[i] = (m02[i] - m20[i]) / S
                qx[i] = (m01[i] + m10[i]) / S
                qy[i] = 0.25 * S
                qz[i] = (m12[i] + m21[i]) / S
            else:
                S = np.sqrt(1.0 + m22[i] - m00[i] - m11[i]) * 2
                qw[i] = (m10[i] - m01[i]) / S
                qx[i] = (m02[i] + m20[i]) / S
                qy[i] = (m12[i] + m21[i]) / S
                qz[i] = 0.25 * S
        outDict['rotation'] = np.concatenate((qw[:, np.newaxis],
                                              qx[:, np.newaxis],
                                              qy[:, np.newaxis],
                                              qz[:, np.newaxis],
                                              ), axis=1)
        del outDict['transformation']
        return outDict

def rotToAxisAngle(inDict):
    rotation = inDict['rotation']
    rotationRepresentation = findRotationRepresentation(rotation)
    if rotationRepresentation == 'axisAngle':
        return inDict
    if rotationRepresentation == 'quaternion':
        raise NotImplementedError("Method not implemented")
    elif rotationRepresentation == 'rVec':
        raise NotImplementedError("Method not implemented")
    else: #rotationRepresentation == 'mat'
        raise NotImplementedError("Method not implemented")

'''
Add into the dict two arrays, containing:
(a) the instantaneous linear speed, calculated as the euclidean distance from
  one point to the next / divided by the time from one point to the next,
  using the pairwise average of previous and next to give an estimate for
  each point.
(b) the instantaneous rotational speed, calculated as above expect for angles
'''
def calculateSpeed(inDict):
    def pairwiseMean(array):
        array = array[:, np.newaxis]
        shiftedArray = np.concatenate((array[1:], array[:-1]), axis=1)
        array = np.mean(shiftedArray, axis=1)
        array = np.insert(array, 0, array[0])
        array = np.append(array, array[-1])
        return array
    linearDiffs = np.linalg.norm(np.diff(inDict['point'], axis=0), axis=1)
    timeDiffs = np.diff(inDict['ts'])
    inDict['linearSpeed'] = pairwiseMean(linearDiffs / timeDiffs)
    inDictAsQuat = rotToRepresentation(inDict, 'quaternion')
    rotations = inDictAsQuat['rotation']
    angularDiffs = angleBetweenTwoQuaternionsArray(rotations[:-1, :], rotations[1:, :])
    inDict['angularSpeed'] = pairwiseMean(angularDiffs / timeDiffs)
    return inDict

def interpolatePoses(inDict, times=None, period=None, maxPeriod=None):
    '''
    method: based on the inputs times, period and maxPeriod,
    create a set of times desired
    crop these outside of the range of the existing times
    Then select original times which are also in the new times
    Then do interpolation for each of the new times
    Then merge the two sets.

    It would be possible to create sub-implementations for each rotation
    representation, but for simplicity we convert to quaternion, do the
    interpolation, and convert back again.

    This function is not compatible with labelled poses, or any other
    non-standard additional fields in the poseDict

    '''
    rotation = inDict['rotation']
    rotationRepresentation = findRotationRepresentation(rotation)
    poseDict = rotToQuat(inDict)

    if np.isscalar(times):
        return interpolatePosesSingle(inDict, time=times)

    ts = poseDict['ts']
    points = poseDict['point']
    rotations = poseDict['rotation']

    firstTime = ts[0]
    lastTime = ts[-1]
    if times is not None:
        keepTimesBool = np.logical_and(times >= firstTime, times <= lastTime)
        times = times[keepTimesBool]
        timesAlreadyExistBool = np.isin(times, ts)
        existingTimes = times[timesAlreadyExistBool]
        newTimes = times[~timesAlreadyExistBool]
    elif period is not None:
        times = np.arange(firstTime, lastTime, step=period)
        timesAlreadyExistBool = np.isin(times, ts)
        existingTimes = times[timesAlreadyExistBool]
        newTimes = times[~timesAlreadyExistBool]
    elif maxPeriod is not None:
        proposedAdditionalTimes = np.arange(firstTime, lastTime, maxPeriod)
        prevIds = np.searchsorted(ts, proposedAdditionalTimes, side='right') - 1
        distPre = proposedAdditionalTimes - ts[prevIds]
        distPost = ts[prevIds + 1] - proposedAdditionalTimes
        dist = (distPre + distPost) * (distPre > 0) * (distPost > 0)
        keepAdditional = dist > maxPeriod
        newTimes = proposedAdditionalTimes[keepAdditional]
        existingTimes = ts

    # Now we have newTimes and existingTimes;
    # Now iterate over newTimes and do interpolation
    numNewTimes = len(newTimes)
    newPoints = np.zeros((numNewTimes, 3), dtype=float)
    newRotations = np.zeros((numNewTimes, 4), dtype=float)
    for idx in range(numNewTimes):
        # TODO: slerp function could be vectorised
        # then this whole loop could be vectorised
        newTime = newTimes[idx]
        idxPre = np.searchsorted(ts, newTime, side='right') - 1
        timePre = ts[idxPre]
        timePost = ts[idxPre + 1]
        rotPre = rotations[idxPre, :]
        rotPost = rotations[idxPre + 1, :]
        timeRel = (newTime - timePre) / (timePost - timePre)
        newRotations[idx, :] = slerp(rotPre, rotPost, timeRel)
        pointPre = points[idxPre, :]
        pointPost = points[idxPre + 1, :]
        newPoints[idx, :] = pointPre * (1-timeRel) + pointPost * timeRel

    # get existingTimes for merging
    if maxPeriod is None:
        keepExistingBool = np.isin(ts, existingTimes)
        ts = ts[keepExistingBool]
        points = points[keepExistingBool, :]
        rotations = rotations[keepExistingBool, :]

    ts = np.concatenate((ts, newTimes))
    points = np.concatenate((points, newPoints))
    rotations = np.concatenate((rotations, newRotations))

    newPoseDict = {}
    newPoseDict['ts'] = ts
    newPoseDict['point'] = points
    newPoseDict['rotation'] = rotations
    for key in poseDict.keys():
        if key not in ['ts', 'point', 'rotation']:
            newPoseDict[key] = poseDict[key].copy()
    newPoseDict = sortDataTypeDictByTime(newPoseDict)
    return rotToRepresentation(newPoseDict, rotationRepresentation)

def rotate(inDict, rotation):
    return transform(inDict, rotation=rotation)

def translate(inDict, translation):
    return transform(inDict, translation=translation)

'''
If you pass in transformationDict, it is applied sample-wise as if the dicts
contained mats and the result were: transformationDict @ poseDict
If instead you pass in transformationMatrix, then this is applied to the whole
array thus:
    transformationMatrix @ poseDict
If instead you pass in either translation or rotation, or both, they are
converted into a transformation matrix, and applied to the whole array.
If 'direction' is 'reverse' then the multiplication is instead applied:
    poseDict @ transformationMatrix

The implementation here first converts to transformation matrices
and then converts back at the end
TODO: There could be efficiency gains by not converting to matrices first
'''
def transform(inDict, transformationDict=None, transformationMatrix=None,
              translation=None, rotation=None, direction='normal'):
    rotationRepresentation = findRotationRepresentation(inDict['rotation'])
    inDict = rotToMat(inDict)
    if transformationDict is not None:
        # Assume that the two dicts have exactly the same timestamps
        # TODO: don't assume this but rather interpolate and merge
        # following interpolatePoses function
        transformationDict = rotToMat(transformationDict)
        transformation = transformationDict['transformation']
    else:
        if transformationMatrix is not None:
            transformation = transformationMatrix
        else:
            # Form a transformation matrix from either or both of translation and rotation
            transformation = np.zeros((4, 4), dtype=float)
            transformation[3, 3] = 1
            if translation is not None:
                transformation[:3, 3] = translation
            if rotation is not None:
                rotRepOfTransformation = findRotationRepresentationSingle(rotation)
                if rotRepOfTransformation == 'quaternion':
                    transformation[:3, :3] = quatToMatSingle(rotation,
                                                M=transformation[:3, :3])
                elif rotRepOfTransformation == 'mat':
                    transformation[:3, :3] = rotation
                else:
                    raise NotImplementedError
    if direction in ['normal', 'forward', 'forwards']:
        transformed = transformation @ inDict['transformation']
    elif direction in ['reverse', 'reversed', 'backward', 'backwards']:
        transformed = inDict['transformation'] @ transformation
    else:
        raise ValueError('"direction" argument not recognised')
    outDict = inDict.copy()
    outDict['point'] = transformed[:, :3, 3]
    outDict['rotation'] = transformed[:, :3, :3]
    outDict['transformation'] = transformed

    return rotToRepresentation(outDict, rotationRepresentation)

# rotation is single
# Assuming that rotation comes in as a quaternion; TODO: don't assume this
def angleBetween(inDict, rotation):
    '''
    if rotation.ndim == 1:
    qP = quaternionProduct(q1, quaternionInverse(q2))
    normV = np.linalg.norm(qP[1:])
    angle = 2 * np.arcsin(normV)
    return outDict
    '''
    raise NotImplementedError('not implemented')

def averageRotation(inDict, weights=None):
    # Convert to quaternion
    inDict = rotToMat(inDict)
    rotations = inDict['rotation']
    numRotations = rotations.shape[0]
    if weights is None:
        weights = np.ones(numRotations)
    A = np.zeros((4,4), dtype=float)
    for weight, q in zip(weights, rotations):
        # multiply q with its transposed version q' and accumulate in A
        A = A + weight * np.outer(q,q)
    # scale
    A = A / numRotations
    # compute eigenvalues and -vectors
    eigenValues, eigenVectors = np.linalg.eig(A)
    # Sort by largest eigenvalue
    eigenVectors = eigenVectors[:,eigenValues.argsort()[::-1]]
    # return the real part of the largest eigenvector (has only real part)
    return eigenVectors[:,0]

"""
A quaternion is in canonical form when it's first component is non-negative.
"""
def canoniseQuaternions(inDict):
    outDict = inDict.copy()
    q = outDict['rotation']
    qw = q[:, 0]
    qSign = qw / np.abs(qw)
    qSign = qSign[:, np.newaxis]
    qCanonised = qSign * q
    outDict['rotation'] = qCanonised
    return outDict

"""
a negative dot product between two quaternions indicates a discontinuity;
pass through the trajectory, flipping at these discontinuities
"""
def makeQuaternionsContinuous(inDict):
    outDict = inDict.copy()
    allQ = outDict['rotation'].copy()
    swapIds = np.where((allQ[:-1, :] * allQ[1:, :]).sum(axis=1) < 0)[0]
    if len(swapIds) % 2 == 1:
        swapIds = np.append(swapIds, (len(allQ) - 1))
    swapBool = np.zeros((len(outDict['ts'])), dtype=bool)

    for swapIdsIdx in range(0, len(swapIds), 2):
        swapBool[swapIds[swapIdsIdx]+1 :swapIds[swapIdsIdx + 1]+1] = True
    allQ[swapBool, :] = - allQ[swapBool, :]
    outDict['rotation'] = allQ
    return outDict

'''
Offsets all points so that the first is at zero.
TODO: think of a good way to store the offset, in reference to tsOffset;
If a np[1,3] array is stored, that will break merge and copy dict functionalities.
'''
def zeroPoint(inDict):
    outDict = inDict.copy()
    outDict['point'] = outDict['point'] - outDict['point'][0, :]
    return outDict

'''
rotate all the poses so that the first pose is at [1w, 0x, 0y, 0z]
then offset
'''
def zeroPose(inDict):
    raise NotImplementedError("Method not implemented")


#%% Better names for legacy functions

def interpolatePosesSingle(poseDict, time):
    ts = poseDict['ts']
    points = poseDict['point']
    rotations = poseDict['rotation']
    idxPre = np.searchsorted(ts, time, side='right') - 1
    timePre = ts[idxPre]
    if timePre == time:
        # In this edge-case of desired time == timestamp, there is no need
        # to interpolate
        return (points[idxPre, :], rotations[idxPre, :])
    if idxPre < 0:
        return (points[0, :], rotations[0, :])
    if idxPre >= len(poseDict['ts']):
        return (points[-1, :], rotations[-1, :])
    timePost = ts[idxPre + 1]
    rotPre = rotations[idxPre, :]
    rotPost = rotations[idxPre + 1, :]
    timeRel = (time - timePre) / (timePost - timePre)
    rotOut = slerp(rotPre, rotPost, timeRel)
    pointPre = points[idxPre, :]
    pointPost = points[idxPre + 1, :]
    pointOut = pointPre * (1-timeRel) + pointPost * timeRel
    return (pointOut, rotOut)

# Can accept an existing matrix, which should be min 3x3;
#if it creates a matrix it makes it 4x4
def quatToMatSingle(quat, M=None):
    if M is None:
        M = np.zeros((4, 4))
        M[3, 3] = 1
    elif M.shape[0] == 4:
        M[:, 3] = 0
        M[3, :] = 0
        M[3, 3] = 1
    w = quat[0]
    x = quat[1]
    y = quat[2]
    z = quat[3]
    M[0, 0] = 1 - 2*y**2 - 2*z**2
    M[0, 1] = 2*x*y - 2*z*w
    M[0, 2] = 2*x*z + 2*y*w
    M[1, 0] = 2*x*y + 2*z*w
    M[1, 1] = 1 - 2*x**2 - 2*z**2
    M[1, 2] = 2*y*z - 2*x*w
    M[2, 0] = 2*x*z - 2*y*w
    M[2, 1] = 2*y*z + 2*x*w
    M[2, 2] = 1 - 2*x**2 - 2*y**2
    return M


'''
# following:
#https://stackoverflow.com/questions/24197182/efficient-quaternion-angular-velocity
def deltaRotationAsQuat(angV):
    vectorOfHalfAngle = angV * 0.5
    norm = np.linalg.norm(vectorOfHalfAngle) # magnitude
    if (norm > 0):
        norm *= np.sin(norm) / norm
        return np.insert(vectorOfHalfAngle, 0, np.cos(norm))
    else:
        return np.insert(vectorOfHalfAngle, 0, 1)
'''


'''
A treatment of instantaneous angular velocities from an imu for the purposes
of visualisation
def angularVelocityToRotMat(angV):
    return linalg.expm(np.array(
        [[0, -angV[2], angV[1]],
         [angV[2], 0, -angV[0]],
         [-angV[1], angV[0], 0]])
'''


def pose6qToMatSingle(pose6q):
    T = pose6q['point']
    quat = pose6q['rotation']
    M = quatToMatSingle(quat)
    M[:3,3] = T
    return M


# Spherical linear interpolation, adapted from https://en.wikipedia.org/wiki/Slerp
DOT_THRESHOLD = 0.9995
def slerp(q1, q2, time_relative):
    dot = np.sum(q1 * q2)
    if dot < 0.0:
        q2 = -q2
        dot = -dot
    if dot > DOT_THRESHOLD:
        result = q1 + time_relative * (q2 - q1)
        return (result.T / np.linalg.norm(result)).T
    theta_0 = np.arccos(dot)
    sin_theta_0 = np.sin(theta_0)
    theta = theta_0 * time_relative
    sin_theta = np.sin(theta)
    s0 = np.cos(theta) - dot * sin_theta / sin_theta_0
    s1 = sin_theta / sin_theta_0
    return (s0 * q1) + (s1 * q2)

'''
expects pose dict in the form: {'ts': 1d np.array of float timestamps,
                                'point': 2d array float of positions [x, y, z],
                                'rotation': 2d array float of quaternions [rw, rx, ry, rz]
                                (i.e. 6dof with rotation as quaternion)}
Two modes of operation:
If time is not None, then returns the interpolated pose at that time -
    returns (point, rotation) tuple, being np.array 1d x 3 and 4 respectively, float,    which is interpolated pose;
Else if maxPeriod is not None, then it returns the entire pose dict,
    but with additional points necessary to ensure that time between samples
    never exceeds maxPeriod
'''

def quaternionProductSingle(q1, q2):
    return np.array([
    q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3],
    q1[0] * q2[1] + q1[1] * q2[0] + q1[2] * q2[3] - q1[3] * q2[2],
    q1[0] * q2[2] - q1[1] * q2[3] + q1[2] * q2[0] + q1[3] * q2[1],
    q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1] + q1[3] * q2[0] ])
    ''' alternative definition - less efficient
    w1 = q1[0]
    v1 = q1[1:4]
    w2 = q2[0]
    v2 = q2[1:4]
    wOut = w1*w2 - np.dot(v1.T, v2)
    vOut = w1*v2 + w2*v1 + np.cross(v1, v2)
    qOut = np.zeros_like(q1)
    qOut[0] = wOut
    qOut[1:4] = vOut
    return qOut
    '''

# Array version of the above
def quaternionProductArray(q1, q2):
    def x(q, idx): # tis function to clean up the following expression
        return q[:, idx, np.newaxis]
    return np.concatenate((
    x(q1,0) * x(q2,0) - x(q1,1) * x(q2,1) - x(q1,2) * x(q2,2) - x(q1,3) * x(q2,3),
    x(q1,0) * x(q2,1) + x(q1,1) * x(q2,0) + x(q1,2) * x(q2,3) - x(q1,3) * x(q2,2),
    x(q1,0) * x(q2,2) - x(q1,1) * x(q2,3) + x(q1,2) * x(q2,0) + x(q1,3) * x(q2,1),
    x(q1,0) * x(q2,3) + x(q1,1) * x(q2,2) - x(q1,2) * x(q2,1) + x(q1,3) * x(q2,0)
    ), axis=1)

def quaternionConjugateSingle(q):
    return np.array([q[0], -q[1], -q[2], -q[3]])

def quaternionConjugateArray(q):
    return np.concatenate((q[:, 0, np.newaxis], -q[:, 1:4]), axis=1)

def quaternionInverseSingle(q):
    return quaternionConjugateSingle(q) / np.sum(q ** 2)

def quaternionInverseArray(q):
    return quaternionConjugateArray(q) / np.sum(q ** 2, axis=1)[:, np.newaxis]

def angleBetweenTwoQuaternionsSingle(q1, q2):
    # returns minimal angle in radians between two unit quaternions, following:
    # https://www.researchgate.net/post/How_do_I_calculate_the_smallest_angle_between_two_quaternions
    qP = quaternionProductSingle(q1, quaternionInverseSingle(q2))
    normV = np.linalg.norm(qP[1:])
    return 2 * np.arcsin(normV)

def angleBetweenTwoQuaternionsArray(q1, q2):
    qP = quaternionProductArray(q1, quaternionInverseArray(q2))
    normV = np.linalg.norm(qP[:, 1:], axis=1)
    return 2 * np.arcsin(normV)

'''
following https://github.com/KieranWynn/pyquaternion/blob/master/pyquaternion/quaternion.py
'''
def axisAngleToQuatSingle(axis, angle):
    mag_sq = np.dot(axis, axis)
    if mag_sq == 0.0:
        raise ZeroDivisionError("Provided rotation axis has no length")
    # Ensure axis is in unit vector form
    if (abs(1.0 - mag_sq) > 1e-12):
        axis = axis / np.sqrt(mag_sq)
    theta = angle / 2.0
    r = np.cos(theta)
    i = axis * np.sin(theta)
    quat = np.concatenate((np.ones((1)) * r, i))
    quat = quat / np.linalg.norm(quat)
    return quat

'''
following: https://answers.unity.com/questions/1266985/convert-rotation-vector-or-matrix-to-quaternion.html
Receives a rotation vector and returns a quaternion
'''
def rVecToQuatSingle(rotV):
    theta = np.linalg.norm(rotV)
    quat = axisAngleToQuatSingle(rotV[:, 0], theta)
    return quat

'''
Expects
    - poseDict in bimvee form {'ts', 'point', 'rotation' as above}
    - translation as np array of x,y,z
    - rotation as a np array of w,x,y,z (quaternion)
If either translation or rotation are passed, these are applied to all poses
in the poseDict.
The rotations are defined wrt local axes, unlike the translations.
Returns a copy of the poseDict, rotated
'''
def transformPosesOld(poseDict, translation=None, rotation=None):
    # Create a copy of the input array - use the same contents
    outDict = {}
    for key in poseDict.keys():
        outDict[key] = poseDict[key]
    if translation is not None:
        outDict['point'] = poseDict['point'] + translation
    if rotation is not None:
        for idx in range(outDict['rotation'].shape[0]): # TODO: this could be matricised
            outDict['rotation'][idx, :] = \
                quaternionProductSingle(outDict['rotation'][idx, :], rotation)
    return outDict


# The following adapted from https://github.com/christophhagen/averaging-quaternions
# Note that the signs of the output quaternion can be reversed,
# since q and -q describe the same orientation.
# w is an optional weight array, which must have the same number of elements
# as the number of quaternions
def averageOfQuaternions(allQ, w=None):
    # Number of quaternions to average
    M = allQ.shape[0]
    A = np.zeros((4,4), dtype = float)
    if w is None:
        w = np.ones(M,)
    weightSum = 0
    for i in range(0, M):
        q = allQ[i, :]
        # multiply q with its transposed version q' and add A
        A = w[i] * np.outer(q,q) + A
        weightSum += w[i]
    # scale
    A = (1.0 / M) * A
    # compute eigenvalues and -vectors
    eigenValues, eigenVectors = np.linalg.eig(A)
    # Sort by largest eigenvalue
    eigenVectors = eigenVectors[:,eigenValues.argsort()[::-1]]
    # return the real part of the largest eigenvector (has only real part)
    return np.real(eigenVectors[:,0])

'''
iitVicon can produce consecutive pose samples with same ts, point and rotation
eliminate these
Ignore different bodyIds for now
'''
def removeDuplicates(poseDict):
    ts = poseDict['ts']
    point = poseDict['point']
    rotation = poseDict['rotation']
    tsDiff = np.diff(ts)
    pointDiff = np.sum(np.diff(point, axis=0), axis=1)
    rotationDiff = np.sum(np.diff(rotation, axis=0), axis=1)
    toKeep = np.logical_or(tsDiff!=0, pointDiff!=0, rotationDiff!=0)
    # remove following elements by shifting toKeep dict forwards
    toKeep = np.insert(toKeep, 0, 'True')
    return selectByBool(poseDict, toKeep)





# -*- coding: utf-8 -*-
"""
Created on Thu Sep 24 11:29:15 2020

@author: sbamford
code contributions from Henri Rebecq
"""


import sys
sys.path.append('/usr/local/lib/python2.7/site-packages')
sys.path.append('/opt/ros/kinetic/lib')
sys.path.append('/opt/ros/kinetic/lib/python2.7/dist-packages')
sys.path.append('/home/sim/catkin_ws/devel/lib/python2.7/dist-packages')
import numpy as np
#import cv2
import rosbag
#from dvs_msgs.msg import Event, EventArray
#from sensor_msgs.msg import CameraInfo
#from sensor_msgs.msg import Image
#from sensor_msgs.msg import Imu
#from geometry_msgs.msg import Vector3
from geometry_msgs.msg import PoseStamped
from cv_bridge import CvBridge #, CvBridgeError
import rospy
from tqdm import tqdm

def exportPose6q(inDict, topicName, bagFile):
    point = inDict['point']
    rotation = inDict['rotation']
    for poseIdx in tqdm(range(inDict['ts'].shape[0])):
        poseMsg = PoseStamped()
        ts = rospy.Time(secs=inDict['ts'][poseIdx])        
        poseMsg.header.stamp = ts
        poseMsg.pose.position.x = point[poseIdx, 0]
        poseMsg.pose.position.y = point[poseIdx, 1]
        poseMsg.pose.position.z = point[poseIdx, 2]
        poseMsg.pose.orientation.w = rotation[poseIdx, 0]
        poseMsg.pose.orientation.x = rotation[poseIdx, 1]
        poseMsg.pose.orientation.y = rotation[poseIdx, 2]
        poseMsg.pose.orientation.z = rotation[poseIdx, 3]
        bagFile.write(topic=topicName, msg=poseMsg, t=ts)

    
def exportFrame(inDict, topicName, bagFile):
    bridge = CvBridge()
    for frameIdx in tqdm(range(inDict['ts'].shape[0])):
        img = inDict['frames'][frameIdx]
        assert img.dtype == np.uint8 # TODO: Handle type conversion here
        # For example by doing right shifts from 10-bit encoding to 8 bits: img = np.right_shift(img, 2)
        img = img.astype('uint8')
        ts = rospy.Time(secs=inDict['ts'][frameIdx])
        imgMsg = bridge.cv2_to_imgmsg(img, 'mono8')
        imgMsg.header.stamp = ts
        bagFile.write(topic=topicName, msg=imgMsg, t=ts)
    
def exportRpgDvsRos(inDict, **kwargs):
    # Open bag
    exportFilePathAndName = kwargs.get('exportFilePathAndName', './')
    bagFile = rosbag.Bag(exportFilePathAndName, 'w')

    #Descend data hierarchy, looking for dataTypes to export
    for channelName in inDict['data']:
        for dataTypeName in inDict['data'][channelName]:
            dataTypeDict = inDict['data'][channelName][dataTypeName]
            topicName = '/' + channelName + '/' + dataTypeName
            if dataTypeName == 'pose6q':
                # For now, we'll make an arbitrary decision which ros message type to use
                exportPose6q(dataTypeDict, topicName, bagFile)
            elif dataTypeName == 'frame':
                # For now, we'll make an arbitrary decision which ros message type to use
                exportFrame(dataTypeDict, topicName, bagFile)
            else:
                print('Skipping dataType "' + dataTypeName + '" from channel "' + channelName + '"')
    bagFile.close()
    
'''
Legacy code from AedatTools repo
#%% DVS
    
    # Put several events into an array in a single ros message, for efficiency     
    
    if 'polarity' in aedat['data'] \
        and ('dataTypes' not in aedat['info'] or 'polarity' in aedat['info']['dataTypes']): 
    legacy code by Henri Rebecq, from AedatTools repo
        countMsgs = 0
        numEventsPerArray = 25000 # Could be a parameter
        numEvents = aedat['data']['polarity']['numEvents']
        numArrays = - (- numEvents / numEventsPerArray) # The subtraction allows rounding up
    
        # Construct the event array object - a definition from rpg_dvs_ros
        # Use this repeatedly for each message        
        eventArrayObject = EventArray()
        # The following properties don't change
        eventArrayObject.width = 240 # HARDCODED CONSTANT - RESOLVE ON A RAINY DAY
        eventArrayObject.height = 180 # HARDCODED CONSTANT - RESOLVE ON A RAINY DAY
        # Use the following object array repeatedly to construct the contents 
        # of each ros message
        eventArray = np.empty(-(-numEventsPerArray), 'object')
        # Outer loop over arrays or ros messages
        for startPointer in range(0, numEvents, numEventsPerArray):         
            countMsgs = countMsgs + 1        
            print 'Writing event array message', countMsgs, 'of', numArrays, ' ...'
            endPointer = min(startPointer + numEventsPerArray, numEvents)            
            # Break the data vectors out of the dict for efficiency, 
            # but do this message by message to avoid memory problems
            arrayX        = aedat['data']['polarity']['x'][startPointer : endPointer] 
            arrayY        = aedat['data']['polarity']['y'][startPointer : endPointer] 
            arrayPolarity = aedat['data']['polarity']['polarity'][startPointer : endPointer]
            # Convert timestamps to seconds (ros, however, stores timestamps to ns precision)            
            arrayTimeStamp = aedat['data']['polarity']['timeStamp'][startPointer : endPointer]/1000000.0 

            # Iterate through all the events in the intended event array
            for eventIndex in range (0, endPointer - startPointer):
                # The Event object definition comes from rpg_dvs_ros
                e = Event()
                e.x = 239 - arrayX[eventIndex] # Flip X - I don't know why this is necessary
                e.y = arrayY[eventIndex]
                e.ts = rospy.Time(arrayTimeStamp[eventIndex])
                e.polarity = arrayPolarity[eventIndex]
                eventArray[eventIndex] = e;
            # The last array may be smaller than numEventsPerArray, so clip the object array
            if endPointer == numEvents:
                eventArray = eventArray[0 : endPointer - startPointer]
            # Assume that the ros message is sent at the time of the last event in the message
            eventArrayObject.header.stamp = e.ts
            eventArrayObject.events = eventArray
            bag.write(topic='/dvs/events', msg=eventArrayObject, t=e.ts)
            
    #%% IMU6
    
    # Put several events into an array in a single ros message, for efficiency     
    if 'imu6' in aedat['data'] \
        and ('dataTypes' not in aedat['info'] or 'imu6' in aedat['info']['dataTypes']): 
        # Break the IMU events out of the dict, for efficiency
        # Accel is imported as g; we want m/s^2
        arrayAccelX = aedat['data']['imu6']['accelX'] * 9.8
        arrayAccelY = aedat['data']['imu6']['accelY'] * 9.8
        arrayAccelZ = aedat['data']['imu6']['accelZ'] * 9.8
        # Angular velocity is imported as deg/s; we want rad/s
        arrayGyroX = aedat['data']['imu6']['gyroX'] * 0.01745
        arrayGyroY = aedat['data']['imu6']['gyroY'] * 0.01745
        arrayGyroZ = aedat['data']['imu6']['gyroZ'] * 0.01745
            # Convert timestamps to seconds (ros, however, stores timestamps to ns precision)            
        arrayTimeStamp = aedat['data']['imu6']['timeStamp']/1000000.0 
        numEvents = aedat['data']['imu6']['numEvents']
        # Use the following containers repeatedly during the export
        imuMsg = Imu()
        accel = Vector3()
        gyro = Vector3()
        # I guess these assignments only need to be made once
        imuMsg.linear_acceleration = accel
        imuMsg.angular_velocity = gyro
        for eventIndex in range(0, numEvents):         
            imuMsg.header.stamp = rospy.Time(arrayTimeStamp[eventIndex])            
            accel.x = arrayAccelX[eventIndex]
            accel.y = arrayAccelY[eventIndex]
            accel.z = arrayAccelZ[eventIndex]
            gyro.x = arrayGyroX[eventIndex]
            gyro.y = arrayGyroY[eventIndex]
            gyro.z = arrayGyroZ[eventIndex]
            bag.write(topic='/dvs/imu', msg=imuMsg, t=imuMsg.header.stamp)
    '''



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Author: Suman Ghosh
        Sim Bamford
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

Input:
vicon data log recorded through yarp

There are two output formats. If 'parameter' separateBodiesAsChannels is present 
and True then the format is:
    outDict = {
        'info': {
            'filePathOrName': <filepath>
            'uniqueIds': np.array of strings, one for each unique bodyId 
            }, 
        'data': { 
            <bodyID>: {
                'pose6q': {
                    'ts' : <1D array of timestamps>,
                    'point' : <2D array where each row has 3d position of a body at a time instant>,
                    'rotation' : <2D array where each row has rotation of a body at a time instant expressed as a quaternion (4d)>} } } }
Otherwise:
    outDict = {
        'info': {
            'filePathOrName': <filepath>
            }, 
        'data': {
            'vicon': {
                'pose6q': {
                    'ts' : <1D array of timestamps>,
                    'point' : <2D array where each row has 3d position of a body at a time instant>,
                    'rotation' : <2D array where each row has rotation of a body at a time instant expressed as a quaternion (4d)>,
                    'bodyId' : <1D array where each row has the bodyId of the corresponding marker>,
                    } } } }
    
A bodyID is the name assigned by vicon to a marker (labeled / unlableled) or rigid body
The pose consists of a point in the form [x, y, z]
and a rotation as a quaternion [r_w, r_x, r_y, r_z] (Caution with alternative orderings here)
The datatype is called 'pose6q', referring to the 6dof with rotation in quaternion form.

Additionally, if separateBodiesAsChannels is not present or false, 
and the separateMarkersAndSegments parameter is present and True,
then the data in the vicon channel is broken into two datatypes:
        ...
            'vicon': {
                'pose6q': {
                    'ts' : <1D array of timestamps>,
                    'point' : <2D array where each row has 3d position of a body at a time instant>,
                    'rotation' : <2D array where each row has the rotation of a body at a time instant expressed as a quaternion (4d)>,
                    'bodyId' : <1D array where each row has the bodyId of the corresponding marker>,
                    } 
                'point3': {
                    'ts' : <1D array of timestamps>,
                    'point' : <2D array where each row has 3d position of a body at a time instant>,
                    'bodyId' : <1D array where each row has the bodyId of the corresponding marker>,
                    } ...
                
Known Issue: timestamps may not be monotonic in the raw data, 
and this function doesn't attempt to correct it.
Timestamps are monotonic for a single segment, but not necessarily for multiple 
segments imported in the same container.
"""

import os
import re
import numpy as np

# local imports
from .timestamps import zeroTimestampsForADataType
from .split import splitByLabel, selectByBool

def getOrInsertDefault(inDict, arg, default):
    # get an arg from a dict.
    # If the the dict doesn't contain the arg, return the default, 
    # and also insert the default into the dict
    value = inDict.get(arg, default)
    if value == default:
        inDict[arg] = default
    return value

# accepts a pose6q datatype dict; returns a channel dict containing pose6q and point3 datatypes
def separateMarkersFromSegments(poseDict):
    isMarker = np.apply_along_axis(lambda x : 'Marker' in str(x[0]), 1, poseDict['bodyId'][..., np.newaxis])
    pointDict = selectByBool(poseDict, isMarker)
    poseDict = selectByBool(poseDict, ~isMarker)
    return {
        'pose6q': poseDict,
        'point3': pointDict}
    
def importIitVicon(**kwargs):
    filePathOrName = kwargs.get('filePathOrName')
    # handle the case in which filename is not specified - iterate through files 
    # in the current directory looking for data.log
    if filePathOrName is None:
        files = [file for file in os.listdir('.') if os.path.isfile(file)]
        for filename in files:
            if filename == 'data.log':
                kwargs['filePathOrName'] = filename
                return importIitVicon(**kwargs)
        print('No suitable file found')
        return None
    pattern = re.compile('(\d+) (\d+\.\d+) \((.*)\)')
    # yarpBottleTimes = []
    outDict = {'info': {'filePathOrName': filePathOrName}, 'data': {}}
    poseDict = {'ts': [], 'point': [], 'rotation': [], 'bodyId': []}
    with open(filePathOrName, 'r') as file:
        print('Found file to read')
        line = file.readline()
        while line:
            found = pattern.search(line.strip())
            # yarpBottleTimes.append(float(found.group(2)))
            viconData = found.group(3)
            bodies = viconData.split(') (')
            for body in bodies:
                elements = body.split(" ")
                bodyId = elements[1].strip('\"')
                ts = elements[2]
                point = elements[3:6]
                # Note: quaternion order is [w,x,y,z] - this is defined by yarp 
                # IFrameTransform component, so ignore vicon documentation
                rotation = elements[6:] 
                poseDict['ts'].append(ts)
                poseDict['point'].append(point)
                poseDict['rotation'].append(rotation)
                poseDict['bodyId'].append(bodyId)
            line = file.readline()

    # converting lists of strings to numpy arrays of objects
    poseDict['ts'] = np.array(poseDict['ts'], dtype=float)
    poseDict['point'] = np.array(poseDict['point'], dtype=float)
    poseDict['rotation'] = np.array(poseDict['rotation'], dtype=float)
    poseDict['bodyId'] = np.array(poseDict['bodyId'], dtype=object)
    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
        zeroTimestampsForADataType(poseDict)
    if kwargs.get('separateBodiesAsChannels', False):
        outDict['info']['uniqueIds'] = np.unique(poseDict['bodyId'])
        separatedBodies = splitByLabel(poseDict, 'bodyId')
        # The next line inserts the missing 'pose6q' dataType level into the hierarchy
        outDict['data'] = {bodyName: {'pose6q': bodyDict} 
                           for bodyName, bodyDict in zip(
                                   separatedBodies.keys(), 
                                   separatedBodies.values())}
    elif kwargs.get('separateMarkersFromSegments', False):
        outDict['data']['vicon'] = separateMarkersFromSegments(poseDict)
    else:            
        outDict['data']['vicon'] = {'pose6q': poseDict}
    return outDict






# -*- coding: utf-8 -*-
'''
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotEventRate takes 'inDict' - a dict containing an imported ae file, 
as created by importAe, and creates a series of plots of event rates.
It creates one plot for each dataType dct contianing a 'ts' field.

Parameters which can be used:
 - min/maxTime
 - min/maxX/Y
 - flipVertical/Horizontal
 - transpose
'''

import numpy as np
import matplotlib.pyplot as plt
from math import log10, floor

def roundToSf(x, sig=3):
    try:
        return round(x, sig-int(floor(log10(abs(x))))-1)
    except ValueError: # log of zero
        return 0

''' 
Descend a container hierarchically, looking for 'ts' fields to work on
'''

def plotEventRate(inDicts, **kwargs):
    if isinstance(inDicts, list):
        for inDict in inDicts:
            plotEventRate(inDict, **kwargs)
        return
    else:
        inDict = inDicts
    if not isinstance(inDict, dict):
        return
    if 'ts' not in inDict:
        title = kwargs.pop('title', '')
        if 'info' in inDict and isinstance(inDict, dict):
            fileName = inDict['info'].get('filePathOrName')
            if fileName is not None:
                print('plotEventRate was called for file ' + fileName)
                title = (title + ' ' + fileName).lstrip()
        for key in inDict.keys():
            kwargs['title'] = (title + ' ' + key).lstrip()
            plotEventRate(inDict[key], **kwargs)
    else: # It's a data-type container
        # Break out data array for cleaner code
        ts = inDict['ts']
        if ts.shape[0] == 0: 
            return
        if ts.shape[0] == 1:
            title = kwargs.get('title', 'this container')
            print('Only one event in ' + title)
            return
        startTime = kwargs.get('startTime', kwargs.get('minTime', kwargs.get('firstTime', np.min(ts))))
        endTime = kwargs.get('endTime', kwargs.get('maxTime', kwargs.get('lastTime', np.max(ts))))
        freqs = kwargs.get('freqs')
        if freqs is None:
            periods = kwargs.get('periods', [0.001, 0.01, 0.1, 1])
        else:
            periods = [1/f for f in freqs]        
        axes = kwargs.get('axes') 
        if axes is None:
            fig, axes = plt.subplots()
            kwargs['axes'] = axes
        legend = kwargs.get('legend', [])
        for period in periods:
            endTimes = [t + period for t in np.arange(startTime, endTime, period)]
            midTimes = [t - period / 2 for t in endTimes]
            endIds = [np.searchsorted(ts, t) for t in endTimes]
            counts = [end-start for start, end in zip(endIds[:-1], endIds[1:])]
            counts.insert(0, endIds[0])
            rates = [count/period for count in counts]
            if kwargs.get('perPixel', False):
                minX = inDict.get('minX', kwargs.get('minX', min(inDict['x']))) 
                maxX = inDict.get('maxX', kwargs.get('maxX', max(inDict['x']))) 
                minY = inDict.get('minY', kwargs.get('minY', min(inDict['y']))) 
                maxY = inDict.get('maxY', kwargs.get('maxY', max(inDict['y']))) 
                numPixels = (maxX - minX + 1) * (maxY - minY + 1)
                rates = rates / numPixels
            endTimes = np.arange(startTime, endTime, period)
            axes.plot(midTimes, rates)
            plt.xlabel('Time (s)')
            plt.ylabel('Rate (events/s)')
            legend.append('period: ' + str(period) + ' s')
        axes.legend(legend)
        kwargs['legend'] = legend
        if kwargs.get('title') is not None:
            axes.set_title(kwargs.get('title'))            
        callback = kwargs.get('callback')
        if callback is not None:
            kwargs = callback(**kwargs)            


                
                


# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)

Using hickle to add hierarchical lists and dicts to hdf5 automatically
https://github.com/telegraphic/hickle
In fact, these are just thin wrappers around hickle.dump/load, 
to offer a similar export function to other export calls.
"""
#%%

import hickle

def importHdf5(filePathOrName='./temp.hdf5', **kwargs):
    #TODO: Handle path with no filename
    if filePathOrName[-5:] != '.hdf5':
        filePathOrName = filePathOrName + '.hdf5'
    print('importHdf5 called, targeting file path and name' + filePathOrName)
    return hickle.load(filePathOrName)


# -*- coding: utf-8 -*-

"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of importRosbag.

"""

#%%

from struct import unpack
import numpy as np

def unpackHeader(headerLen, headerBytes):
    fields = {}
    ptr = 0
    while ptr < headerLen:
        fieldLen = unpack('=l', headerBytes[ptr:ptr+4])[0]
        ptr += 4
        #print(fieldLen)
        field = headerBytes[ptr:ptr+fieldLen]
        ptr += fieldLen
        #print(field)
        fieldSplit = field.find(b'\x3d')
        fieldName = field[:fieldSplit].decode("utf-8")
        fieldValue = field[fieldSplit+1:]
        fields[fieldName] = fieldValue
    return fields

def unpackRosUint32(data, ptr):
    return unpack('=L', data[ptr:ptr+4])[0], ptr+4

def unpackRosUint8(data, ptr):
    return unpack('=B', data[ptr:ptr+1])[0], ptr+1

def unpackRosString(data, ptr):
    stringLen = unpack('=L', data[ptr:ptr+4])[0]
    ptr += 4
    try:
        outStr = data[ptr:ptr+stringLen].decode('utf-8')
    except UnicodeDecodeError:
        outStr = 'UnicodeDecodeError'
    ptr += stringLen
    return outStr, ptr

def unpackRosFloat64Array(data, num, ptr):
    return np.frombuffer(data[ptr:ptr+num*8], dtype=np.float64), ptr+num*8 
    
def unpackRosFloat32Array(data, num, ptr):
    return np.frombuffer(data[ptr:ptr+num*4], dtype=np.float32), ptr+num*4 
    
def unpackRosFloat32(data, ptr):
    return unpack('=f', data[ptr:ptr+4])[0], ptr+4 
    
def unpackRosTimestamp(data, ptr):
    timeS, timeNs = unpack('=LL', data[ptr:ptr+8])
    timeFloat = np.float64(timeS)+np.float64(timeNs)*0.000000001 
    return timeFloat, ptr+8



# -*- coding: utf-8 -*-

"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of importRosbag.

The importTopic function receives a list of messages and returns
a dict with one field for each data field in the message, where the field
will contain an appropriate iterable to contain the interpretted contents of each message.
In some cases, static info is repeated in each message; in which case a field may not contain an iterable. 

This function imports the ros message type defined at:
https://github.com/uzh-rpg/rpg_dvs_ros/tree/master/dvs_msgs/msg
"""

#%%

from tqdm import tqdm
import numpy as np

# local imports

from .common import unpackRosString, unpackRosUint32

def importTopic(msgs, **kwargs):

    disable_bar = kwargs.get('disable_bar')
    tsByMessage = []
    xByMessage = []
    yByMessage = []
    polByMessage = []
    for msg in tqdm(msgs, position=0, leave=True, disable=disable_bar):
        # TODO: maybe implement kwargs['useRosMsgTimestamps']
        data = msg['data']
        #seq = unpack('=L', data[0:4])[0]
        #timeS, timeNs = unpack('=LL', data[4:12])
        frame_id, ptr = unpackRosString(data, 12)
        height, ptr = unpackRosUint32(data, ptr)
        width, ptr = unpackRosUint32(data, ptr) 
        numEventsInMsg, ptr = unpackRosUint32(data, ptr)
        # The format of the event is x=Uint16, y=Uint16, ts = Uint32, tns (nano seconds) = Uint32, pol=Bool  
        # Unpack in batch into uint8 and then compose
        dataAsArray = np.frombuffer(data[ptr:ptr+numEventsInMsg*13], dtype=np.uint8)
        dataAsArray = dataAsArray.reshape((-1, 13), order='C')
        # Assuming big-endian
        xByMessage.append((dataAsArray[:, 0] + dataAsArray[:, 1] * 2**8).astype(np.uint16))
        yByMessage.append((dataAsArray[:, 2] + dataAsArray[:, 3] * 2**8).astype(np.uint16))
        ts = ((dataAsArray[:, 4] + \
              dataAsArray[:, 5] * 2**8 + \
              dataAsArray[:, 6] * 2**16 + \
              dataAsArray[:, 7] * 2**24 ).astype(np.float64))
        tns = ((dataAsArray[:, 8] + \
               dataAsArray[:, 9] * 2**8 + \
               dataAsArray[:, 10] * 2**16 + \
               dataAsArray[:, 11] * 2**24).astype(np.float64))
        tsByMessage.append(ts + tns / 1000000000) # Combine timestamp parts, result is in seconds
        polByMessage.append(dataAsArray[:, 12].astype(bool))
    outDict = {
        'x': np.concatenate(xByMessage),
        'y': np.concatenate(yByMessage),
        'ts': np.concatenate(tsByMessage),
        'pol': np.concatenate(polByMessage),
        'dimX': width,
        'dimY': height}
    return outDict



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Author: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotCorrelogram receives "inDict", which is a container for (2D) polarity events, 
(Or a higher level container), containing fields with numpy arrays:    
    ts
    x
    y
    pol
Autocorrelation
A correlogram is created for the chosen range of pixels.
Choose the pixels by setting the kwargs:
    xMin
    xMax
    yMin
    yMax
By default, all pixels are included. 
By default, all pixels are autocorrelated individually, and then the results are 
combined, so the end result is a mean over all pixels.
TODO: Could do an autocorrelation for all spikes for the range together.
minTime and maxTime are used to constrain time. 
By default, polarities are combined; separate them with splitByPol=true
numBins is interpretted as the number of time bins in each of positive 
and negative directions, not including a central bin. 
"""

#%% Plot tsPix (single pixel timestamps)

import matplotlib.pyplot as plt
import numpy as np
from tqdm import trange

# local imports
from .split import cropSpace, cropTime, splitByPolarity
    
# TODO: include a linear spacing option
def defineBoundariesAndDensities(**kwargs):
    numBins = kwargs.get('numBins', 50)
    minOffset = kwargs.get('minOffset', 0.00001) # 10 us
    maxOffset = kwargs.get('maxOffset', 1.0) # 100 ms
    logSpacing = kwargs.get('spacing', 'linear') == 'log'
    mirrored = kwargs.get('mirrored', True)
    if logSpacing:
        boundaries = np.geomspace(minOffset, maxOffset, numBins+1)
        if mirrored:
            boundaries = np.concatenate((-np.flip(boundaries), boundaries))
        elif not mirrored:
            boundaries = np.concatenate((np.zeros((1)), boundaries))        
    elif (not logSpacing) and mirrored: # Linear; ignore minOffset
        boundaries = np.linspace(-maxOffset, maxOffset, numBins*2)
    elif (not logSpacing) and (not mirrored):
        boundaries = np.linspace(0, maxOffset, (numBins))
    # now there are boundaries either starting from zero ot mirrored around zero 
    # for each bin, including a central bin, and the extreme boundaries 
    # are the edge of the time region of interest.
    widths = (boundaries[1:] - boundaries[:-1])
    densities = 1 / widths
    centres = (boundaries[1:] + boundaries[:-1]) / 2
    return boundaries, densities, centres, widths

def interSpikeInterval(ts, boundaries, densities, **kwargs):
    hist = np.zeros((len(densities)), dtype=float)
    diff = ts[1:] - ts[:-1]
    for idx, (boundary1, boundary2) in enumerate(zip(boundaries[:-1], boundaries[1:])):
        hist[idx] = np.sum(np.logical_and(diff >= boundary1, diff <= boundary2))
    hist = hist * densities
    sumHist = np.sum(hist) # Avoid dividing by zero in case the hist is empty
    if sumHist:
        hist = hist / sumHist
    return hist

# Plot the timings of the spikes in events2 w.r.t. the spikes in events1
def plotInterSpikeIntervalSingle(events, **kwargs):    
    if (len(events['ts']) == 0):
        return
    # use other library functions to do spatiotemporal cropping
    # If this is called from plotInterSpikeInterval then this has already been done, 
    # But do it here as well so that this function can be independent.
    events = cropSpace(events, **kwargs)
    events = cropTime(events, zeroTime=False, **kwargs)
    kwargs['mirrored'] = False
    boundaries, densities, centres, widths = defineBoundariesAndDensities(**kwargs)
    # Spatial cropping has happened; now iterate through defacto spatial range
    minX = np.min(events['x'])
    maxX = np.max(events['x'])
    minY = np.min(events['y'])
    maxY = np.max(events['y'])
    hist = np.zeros((len(densities)), dtype=float)
    for currX in trange(minX, maxX+1, leave=True, position=0):
        currXLogical = events['x']==currX
        tsForCurrX = events['ts'][currXLogical]
        yForCurrX = events['y'][currXLogical]
        for currY in range(minY, maxY+1):
            tsXY = tsForCurrX[yForCurrX==currY]
            if np.any(tsXY): 
                hist = hist + interSpikeInterval(tsXY, boundaries, densities, **kwargs)
    # Normalise
    weightedSum = np.sum(hist / densities)
    if weightedSum > 0:
        hist = hist / np.sum(weightedSum)
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
        kwargs['axes'] = axes

    axes.bar(centres, hist, widths, antialiased=False)
    #axes.set_ylim(minX + minY*numX - 1, maxX + maxY*numX + 1)
    #plt.xticks(theRange, labels)
    title = kwargs.get('title', '')
    if kwargs.get('minX', False) or  kwargs.get('maxX', False):
        title = title + ' ' + str(minX) + '<=X<=' + str(maxX)
    if kwargs.get('minY', False) or  kwargs.get('maxY', False):
        title = title + ' ' + str(minY) + '<=Y<=' + str(maxY)
    axes.set_title(title)

    callback = kwargs.get('callback')
    if callback is not None:
        callback(**kwargs) 
        
def plotInterSpikeInterval(inDict, **kwargs):
    # Boilerplate for descending container hierarchy
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plotInterSpikeInterval(inDictInst, **kwargs)
        return
    if 'info' in inDict: # Top level container
        fileName = inDict['info'].get('filePathOrName', '')
        print('plotInterSpikeInterval was called for file ' + fileName)
        if not inDict['data']:
            print('The import contains no data.')
            return
        for channelName in inDict['data']:
            channelData = inDict['data'][channelName]
            if 'dvs' in channelData and len(channelData['dvs']['ts']) > 0:
                kwargs['title'] = ' '.join([fileName, str(channelName)])
                plotInterSpikeInterval(channelData['dvs'], **kwargs)
            else:
                print('Channel ' + channelName + ' skipped because it contains no polarity data')
        return
    print('plotInterSpikeInterval working: ' + kwargs.get('title', 'unnamed'))
    # use other library functions to do spatiotemporal cropping
    inDict = cropSpace(inDict, **kwargs)
    inDict = cropTime(inDict, zeroTime=False, **kwargs)
    if kwargs.get('splitByPol', False):
        splitDict = splitByPolarity(inDict)        
        fig, axes = plt.subplots(1, 3)

        kwargs['axes'] = axes[0]
        kwargs['title'] = 'on + off'
        print(kwargs['title'])
        plotInterSpikeIntervalSingle(inDict, **kwargs)

        kwargs['axes'] = axes[1]
        kwargs['title'] = 'off'
        print(kwargs['title'])
        plotInterSpikeIntervalSingle(splitDict['0'], **kwargs)

        kwargs['axes'] = axes[2]
        kwargs['title'] = 'on'
        print(kwargs['title'])
        plotInterSpikeIntervalSingle(splitDict['1'], **kwargs)

    else:            
        axes = kwargs.get('axes')
        if axes is None:
            fig, axes = plt.subplots()
            kwargs['axes'] = axes
        kwargs['title'] = 'auto'
        plotInterSpikeIntervalSingle(inDict, **kwargs)
    
# Alias using common abbreviation
def plotIsi(inDict, **kwargs):
    plotInterSpikeInterval(inDict, **kwargs)

def crossCorrelation(ts1, ts2, boundaries, densities, **kwargs):
    hist = np.zeros((len(densities)), dtype=float)
    # TODO: think about a nice array-based way to do the following:
    for ts in ts1:
        firstIdx = np.searchsorted(ts2, ts+boundaries[0])
        lastIdx = np.searchsorted(ts2, ts+boundaries[-1])
        selectedTs2 = ts2[firstIdx: lastIdx] - ts
        histIds = np.searchsorted(boundaries, selectedTs2) - 1
        for histIdx in histIds:
            try:
                hist[histIdx] = hist[histIdx] + 1
            except IndexError:
                pass # Out of bounds - we don't want it
    # If this was intended as an autocorrelation then we need to remove the 
    # self-correlations
    if kwargs.get('auto', False):
        centreIdx = int(len(boundaries) / 2) - 1 
        hist[centreIdx] = hist[centreIdx] - len(ts1)
    hist = hist * densities
    sumHist = np.sum(hist) # Avoid dividing by zero in case the hist is empty
    if sumHist:
        hist = hist / sumHist
    return hist

# Plot the timings of the spikes in events2 w.r.t. the spikes in events1
def plotCrossCorrelation(events1, events2, **kwargs):    
    if (len(events1['ts']) == 0) or (len(events2['ts']) == 0):
        return
    # use other library functions to do spatiotemporal cropping
    # If this is called from plotautocorrelation then this has already been done, 
    # But do it here as well so that this function can be independent.
    events1 = cropSpace(events1, **kwargs)
    events1 = cropTime(events1, zeroTime=False, **kwargs)
    events2 = cropSpace(events2, **kwargs)
    events2 = cropTime(events2, zeroTime=False, **kwargs)
    boundaries, densities, centres, widths = defineBoundariesAndDensities(**kwargs)
    # Spatial cropping has happened; now iterate through defacto spatial range
    minX = min(np.min(events1['x']), np.min(events2['x']))
    maxX = max(np.max(events1['x']), np.max(events2['x']))
    minY = min(np.min(events1['y']), np.min(events2['y']))
    maxY = max(np.max(events1['y']), np.max(events2['y']))
    hist = np.zeros((len(densities)), dtype=float)
    # TODO: This loop could be much more efficient, following the pattern in plotInterSpikeInterval, above
    for currX in trange(minX, maxX+1, leave=True, position=0): 
        for currY in range(minY, maxY+1):
            ts1 = events1['ts'][np.logical_and(events1['x']==currX, events1['y']==currY)]
            ts2 = events2['ts'][np.logical_and(events2['x']==currX, events2['y']==currY)]
            if np.any(ts1) and np.any(ts2): 
                hist = hist + crossCorrelation(ts1, ts2, boundaries, densities, **kwargs)
    # Normalise
    weightedSum = np.sum(hist / densities)
    if weightedSum > 0:
        hist = hist / np.sum(weightedSum)
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
        kwargs['axes'] = axes

    axes.bar(centres, hist, widths, antialiased=False)
    #axes.set_ylim(minX + minY*numX - 1, maxX + maxY*numX + 1)
    #plt.xticks(theRange, labels)
    title = kwargs.get('title', '')
    if kwargs.get('minX', False) or  kwargs.get('maxX', False):
        title = title + ' ' + str(minX) + '<=X<=' + str(maxX)
    if kwargs.get('minY', False) or  kwargs.get('maxY', False):
        title = title + ' ' + str(minY) + '<=Y<=' + str(maxY)
    axes.set_title(title)

    callback = kwargs.get('callback')
    if callback is not None:
        callback(**kwargs)    

def plotAutoCorrelation(inDict, **kwargs):
    # Boilerplate for descending container hierarchy
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plotCorrelogram(inDictInst, **kwargs)
        return
    if 'info' in inDict: # Top level container
        fileName = inDict['info'].get('filePathOrName', '')
        print('plotAutocorrelation was called for file ' + fileName)
        if not inDict['data']:
            print('The import contains no data.')
            return
        for channelName in inDict['data']:
            channelData = inDict['data'][channelName]
            if 'dvs' in channelData and len(channelData['dvs']['ts']) > 0:
                kwargs['title'] = ' '.join([fileName, str(channelName)])
                plotCorrelogram(channelData['dvs'], **kwargs)
            else:
                print('Channel ' + channelName + ' skipped because it contains no polarity data')
        return
    print('plotAutocorrelation working: ' + kwargs.get('title', 'unnamed'))
    # use other library functions to do spatiotemporal cropping
    inDict = cropSpace(inDict, **kwargs)
    inDict = cropTime(inDict, zeroTime=False, **kwargs)
    if kwargs.get('splitByPol', False):
        splitDict = splitByPolarity(inDict)        
        fig, axes = plt.subplots(2, 2)

        kwargs['axes'] = axes[0, 0]
        kwargs['title'] = 'on + off auto'
        print(kwargs['title'])
        kwargs['auto'] = True
        plotCrossCorrelation(inDict, inDict, **kwargs)

        kwargs['axes'] = axes[0, 1]
        kwargs['title'] = 'off auto'
        print(kwargs['title'])
        kwargs['auto'] = True
        plotCrossCorrelation(splitDict['0'], splitDict['0'], **kwargs)

        kwargs['axes'] = axes[1, 0]
        kwargs['title'] = 'on auto'
        print(kwargs['title'])
        kwargs['auto'] = True
        plotCrossCorrelation(splitDict['1'], splitDict['1'], **kwargs)

        kwargs['axes'] = axes[1, 1]
        kwargs['title'] = 'on w.r.t. off cross'
        print(kwargs['title'])
        kwargs['auto'] = False
        plotCrossCorrelation(splitDict['0'], splitDict['1'], **kwargs)

    elif kwargs.get('splitByPolAlt', False):
        splitDict = splitByPolarity(inDict)        
        fig, axes = plt.subplots(3, 1)
        kwargs['axes'] = axes[0]
        kwargs['title'] = 'on + off auto'
        print(kwargs['title'])
        kwargs['auto'] = True
        plotCrossCorrelation(inDict, inDict, **kwargs)
        kwargs['axes'] = axes[1]
        kwargs['title'] = 'on w.r.t. off cross'
        print(kwargs['title'])
        kwargs['auto'] = False
        plotCrossCorrelation(splitDict['0'], splitDict['1'], **kwargs)
        kwargs['axes'] = axes[2]
        kwargs['title'] = 'off w.r.t. on cross'
        print(kwargs['title'])
        kwargs['auto'] = False
        plotCrossCorrelation(splitDict['1'], splitDict['0'], **kwargs)
    else:            
        axes = kwargs.get('axes')
        if axes is None:
            fig, axes = plt.subplots()
            kwargs['axes'] = axes
        kwargs['title'] = 'auto'
        kwargs['auto'] = True
        plotCrossCorrelation(inDict, inDict, **kwargs)

# For a single trace, this is synonymous with autocorrelation
def plotCorrelogram(inDict, **kwargs):
    plotAutoCorrelation(inDict, **kwargs)


#%%
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: adinale
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Aiko Dinale
        Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
importExportBatches... is for importing timestamped address-event data, 
and then exporting it again possibly in a different format. 
Intended for use with large files, it allows you to read in a certain amount 
of data in a batch, in order not to exceed the limits of working memory.
At the time of writing there is only a function:
    
- importExportBatchesIitYarp
    
Ideally we will extend this to handle all formats.     

Use params numberBatches and importMaxBytes for batch control. the info field
in the imported dict contains running info about the state of the import, 
but this in not encoded in the output file.
"""

import os
from .importIitYarp import importIitYarp
from .exportIitYarp import exportIitYarp
from .timestamps import offsetTimestampsForAContainer

def importExportBatchesIitYarp(datasetPath, numberBatches, importMaxBytes = 1000000, tsBits = 30):
    """
    Import a binary dataset into batches, then export each single batch into YARP format.

    Args:
        datasetPath (string): path containing the dataset in binary format, i.e. binaryevents.log
        numberBatches (int): number of batches which the input dataset should be split into
        importMaxBytes (int, optional): maximum numbe of Bytes to be imported. Defaults to 1000000 (1 MB).
    """    
    dataSizeBytes = os.path.getsize(datasetPath + "/binaryevents.log")

    numMegaByte = round(dataSizeBytes/importMaxBytes)
    sizeBatchMegaByte = round(numMegaByte/numberBatches)
    importedToByte = -1

    for idx in range(0, numberBatches):
        wrapOffset = 0
        for megaByteCounter in range(0, sizeBatchMegaByte):

            importedBatch = importIitYarp(filePathOrName = datasetPath,
                                        tsBits = tsBits,
                                        convertSamplesToImu = False,
                                        importFromByte = importedToByte + 1,
                                        importMaxBytes = importMaxBytes)

            if megaByteCounter == 0:
                # The first batch is treated differently
                exportIitYarp(importedBatch,
                            exportFilePath = os.path.join(datasetPath + "_convertedBatch" + str(idx)),
                            pathForPlayback =  os.path.join(datasetPath + "_convertedBatch" + str(idx)),
                            dataTypes = ['sample', 'dvs'],
                            protectedWrite = False,
                            writeMode = 'w')
                first_tsOffset = importedBatch['info']['tsOffsetFromData']
                previous_tsOffset = first_tsOffset
            else:
                # Offset timestamps in the second batch
                imported_tsOffset = importedBatch['info']['tsOffsetFromData']
                if imported_tsOffset > previous_tsOffset:
                    wrapOffset += 2**tsBits * 0.08 / 1000000

                offsetToApplyToImportedBatch = first_tsOffset - imported_tsOffset + wrapOffset
                offsetTimestampsForAContainer(importedBatch, offsetToApplyToImportedBatch)
                importedBatch['info']['tsOffsetFromData'] += offsetToApplyToImportedBatch

                exportIitYarp(importedBatch,
                            exportFilePath = os.path.join(datasetPath + "_convertedBatch" + str(idx)),
                            pathForPlayback =  os.path.join(datasetPath + "_convertedBatch" + str(idx)),
                            dataTypes = ['sample', 'dvs'],
                            protectedWrite = False,
                            writeMode = 'a')

                previous_tsOffset = imported_tsOffset

            time = importedBatch['data']['right']['dvs']['ts'][-1]
            print("Time " + str(time))

            importedToByte = importedBatch['info']['importedToByte']
            print(str(megaByteCounter) + ": imported to Byte " + str(importedToByte) + "\n")



# -*- coding: utf-8 -*-
"""
Copyright (C) 2021 Event-driven Perception for Robotics
Authors: Simon Muller-Cleve
         Sim Bamford
         
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation
and Export of Events etc)

the samplesToEvents function provides generic conversion from dense single- or
multi-channel data samples (which have typically but not necessarily been
sampled regularly in time) to polarised address events.

inDict comes in containing at least: 
'ts' - float64, assumed monotonically increasing
'value unless otherwise named by the kwarg 'valueKey' - a numpy array of samples
with one row (zeroth dimension) for each ts, and additional dimensions sufficient
to contain the address space. For example, if the input is from a camera,
there may be two additional dimensions for greyscale or 3 in case of rgb data.
Hereafter, a single pixel / taxel etc is referred to as a sensor.

The value data is assumed to be treated in the linear scale - i.e. threshold
applied linearly.
If instead relative thresholding is desired, then for the most general treatment,
the samples should be converted to the log domain prior to calling this function.
TODO: this could be internalised.

For each sensor, the following treatment should be performed.
Sample the first value. Move forward, applying any drift, until a threshold
crossing is found, taking into consideration any drift value.
Commit an event of the appropriate type at the appropriate linearly interpolated
time. Then move forward by the refractory period, before resampling the signal
at that moment. Repeat.
TODO: drift not yet implemented.

TODO: ...
    drift and refractory periods may be randomised by sensor.
    additive noise might be an option, perhaps following v2e, but need
    to think about generality of noise model.

Note: The code is somewhat inefficient in that it doesn't make use of array
operations across multiple sensors. On the other hand, it's probably a better
starting point for a Cuda implementation, providing only that we are genuinely
converting data in batch.
'''
"""

#%%

import numpy as np

from .split import mergeDataTypeDicts

def samplesToEvents(inDict, **kwargs):
    ts = inDict['ts']
    numSamples = len(ts)
    valueKey = kwargs.pop('valueKey', 'value')
    values = inDict[valueKey]
    assert values.shape[0] == numSamples
    ndims = values.ndim
    if ndims > 1:
        # Iterate through the first dimension of values
        results = []
        for addr in range(values.shape[1]):
            subValues = np.take(values, addr, axis=1)
            subValues = np.squeeze(subValues)
            subInDict = {'ts': ts,
                         'value': subValues}
            subOutDict = samplesToEvents(subInDict, **kwargs)
            numSubSamples = len(subOutDict['ts'])
            subAddrs = subOutDict.pop('addr', None)
            newAddrs = np.ones((numSubSamples), dtype = int) * addr
            if subAddrs is None:
                subOutDict['addr'] = newAddrs
            else:
                subOutDict['addr'] = np.concatenate((newAddrs, subAddrs), axis=1)
            results.append(subOutDict)
    else:
        # We have a single 'sensor' - get on with the conversion
        # Unpack params
        thr_ON = kwargs.get('thresholdUp', kwargs.get('threshold', None))
        if thr_ON is None:
            raise NotImplementedError('Choose a method for setting the thresholds')
        thr_OFF = -kwargs.get('thresholdDown', thr_ON) #Note: should be defined positively in the first case
        refr_per = kwargs.get('refractoryPeriod', 0.0)
        # build lists for new events, before converting to arrays at the end
        t_event = ts[0] - refr_per
        last_val = values[0]
        ON_spikes = []
        OFF_spikes = []
        values = values.astype(float) # TODO: consider if this is necessary
        delta_input = np.diff(values)
        dt = np.diff(ts)
        m = delta_input/dt
    
        # calc first event in new sample
        first_event_sample = False
        # create dummy to check if in sample or above
        t_event_dummy = None

        for i in range(len(delta_input)):
            # for the time between two samples
            # calc in between two samples
            in_sample = True
    
            while in_sample:
                # calculate after change from one sample to the next
                if first_event_sample:
                    # ON event
                    if delta_input[i] > 0:
                        # thr reached after refr. per.
                        if ((values[i]+m[i]*(t_event + refr_per - ts[i])) - 
                            last_val >= thr_ON):
                            t_event_dummy = t_event + refr_per
                            # check if t_event in sample or above
                            if t_event_dummy > ts[i + 1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                t_event = t_event_dummy
                                last_val = values[i] + m[i]*(t_event - ts[i])
                                ON_spikes.append(t_event)
                                first_event_sample = False
                        # after refe. per. thr. not reached
                        else:
                            t_event_dummy = (((last_val + thr_ON - values[i])
                                              / m[i]) + ts[i])
                            # check if t_event in sample or above
                            if t_event_dummy > ts[i + 1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                last_val = last_val + thr_ON
                                t_event = t_event_dummy
                                ON_spikes.append(t_event)
                                first_event_sample = False
                    # OFF event
                    elif delta_input[i] < 0:
                        # thr reached after refr. per.
                        if ((values[i] + m[i] * (t_event + refr_per - ts[i]))
                           -last_val <= thr_OFF):
                            t_event_dummy = t_event + refr_per
                            # check if t_event in sample or above
                            if t_event_dummy > ts[i+1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                t_event = t_event_dummy
                                last_val = values[i] + m[i] * (t_event - ts[i])
                                OFF_spikes.append(t_event)
                                first_event_sample = False
                        # after refe. per. thr. not reached
                        else:
                            t_event_dummy = (((last_val + thr_OFF - values[i])
                                             / m[i]) + ts[i])
                            # check if t_event in sample or above
                            if t_event_dummy > ts[i+1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                last_val = last_val + thr_OFF
                                t_event = t_event_dummy
                                OFF_spikes.append(t_event)
                                first_event_sample = False
                    # no change in this sample
                    else:
                        in_sample = False
                        first_event_sample = True
                # calculate events in sample
                else:
                    # ON events
                    if delta_input[i] > 0:
                        # thr. reached after ref. per.
                        if refr_per*m[i] >= thr_ON:
                            t_event_dummy = t_event + refr_per
                            # check if t_event in sample or above
                            if t_event_dummy > ts[i+1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                t_event = t_event_dummy
                                last_val = refr_per*m[i] + last_val
                                ON_spikes.append(t_event)
                        # thr. not reached after ref.
                        else:
                            t_event_dummy = (thr_ON/m[i]) + t_event
                            # check if event is in sample or above
                            if t_event_dummy > ts[i+1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                t_event = t_event_dummy
                                last_val = last_val + thr_ON
                                ON_spikes.append(t_event)
                    # OFF events
                    elif delta_input[i] < 0:
                        # thr. reached after ref. per.
                        if refr_per*m[i] <= thr_OFF:
                            t_event_dummy = t_event + refr_per
                            # check if t_event in sample or above
                            if t_event_dummy > ts[i+1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                t_event = t_event_dummy
                                last_val = refr_per*m[i] + last_val
                                OFF_spikes.append(t_event)
                        # thr. not reached after ref.
                        else:
                            t_event_dummy = (thr_OFF/m[i]) + t_event
                            # check if event is in sample or above
                            if t_event_dummy > ts[i+1]:
                                in_sample = False
                                first_event_sample = True
                            else:
                                t_event = t_event_dummy
                                last_val = last_val + thr_OFF
                                OFF_spikes.append(t_event)
                    # no change in this sample
                    else:
                        in_sample = False
                        # TODO double check here!
                        first_event_sample = True
        # Now we have lists of ON and OFF times - convert to a single sorted
        #dataType dict, including a 'pol' field
        ON_spikes = np.array(ON_spikes, dtype = float)
        onDict = {'ts': ON_spikes,
                  'pol': np.ones_like(ON_spikes, dtype=bool)
                  }
        OFF_spikes = np.array(OFF_spikes, dtype = float)
        offDict = {'ts': OFF_spikes,
                  'pol': np.zeros_like(OFF_spikes, dtype=bool)
                  }
        results = [onDict, offDict]
    return mergeDataTypeDicts(results) # This will sort by default

#%%




# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
         Aiko Dinale
         Massimiliano Iacono
Code contributions from:
        Marco Monforte
        Daniel Gutierrez Galan
        Ali Dabbous
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
`importIitYarp` opens a `data.log` file, initially assumed to contain 24-bit encoded data from ATIS.
Returns a list of dicts, where each dict contains:
{'info': {},
 'data': {
         channel0: {} (usually the 'left' channel)
         channel1: {} (usually the 'right' channel)
         ...
         }}

Each channel is a dict containing:
    For dvs:
        - "ts": numpy array of float - seconds
        - "x": numpy array of uint16
        - "y": numpy array of uint16
        - "pol": numpy array of uint8 in [0, 1]
    For imu:
        - "ts": numpy array of float - seconds

    Can also handle LAE (labelled events), FLOW (dvs optical flow events),
    IMU (imu samples), SKE (skin events), EAR (auditory events), SKS (skin samples)

This function combines the import of several distinctly different cases:
    1) Output directly from zynqGrabber, with all events together.
        May combine samples and events in the same bottles and never contains
        explicit IMU / LAE / FLOW events. Samples such as IMU are contained
        within AE type bottles.
    2) Split and processed events from vPreProcess or other downstream modules.
        The latter usually has distinct bottles for each datatype, and channels
        have probably, but not definitely, been split into left and right.
    3) Output from binary event dumper.
    4) Data from Vicon.
    5) Skin samples which have not been properly formatted as bottles.

TODO: import frames - yarp has 2 ways of encoding - as bottles or,
recommended, by saving e.g. pngs to a folder - support the latter

There are 2 dvs codecs: 20-bit and 24-bit. 24-bit is assumed unless kwarg
"codec" = "20bit" is received

More information about event decoding are available in the event-driven reposi-
tory on GitHub:
    https://github.com/robotology/event-driven/blob/master/documentation/eventcodecs.md
"""

# %%

import re
import numpy as np
import os
import csv

# Local imports
from .importIitVicon import importIitVicon
from .timestamps import unwrapTimestamps, zeroTimestampsForAChannel, rezeroTimestampsForImportedDicts
from .split import selectByLabel
from bimvee.importBoundingBoxes import importBoundingBoxes
from bimvee.importSkeleton import importSkeleton
from tqdm import tqdm

def decodeEvents(data, **kwargs):
    """
    Decode the binary patterns in `data` to the corresponding event type.

    Timestamps (ts) are 32 bit integers counted with an 80 ns clock - do this conversion later.
    There is then a 32 bit data payload.
    The following patterns mark the different data types (MSB left - LSB right):
        0000 0XXX XXXX XXXX: dvs
        0000 1XXX XXXX XXXX: aps
        0001 0XXX XXXX XXXX: skin
        0010 0XXX XXXX XXXX: imu
        0100 0XXX XXXX XXXX: ear
    DVS events are encoded as 32 bits with x, y, channel (ch)(c) and polarity (pol)(p)
    in two possible configuration as shown below:
        0000 0000 000c 00yy yyyy yyxx xxxx xxxp ("20-bit codec" - this codec is now deprecated but useful to load old datasets)
        0000 0000 0crr yyyy yyyy rrxx xxxx xxxp ("24 bit codec")
    ... where:
            - x and y = respective coords;
            - p = polarity;
            - r = reserved for future expansion, but practically used for larger sensor formats
            - c = channel 1 = right vs 0 = left.
    APS:
        not written yet
    IMU samples:
        rrrr rrrr tcrr ssss vvvv vvvv vvvv vvvv
        r = reserved = 0
        t = type: 1=sample
        c = channel (left vs right)
        s = sensor (uint4)
        v = value of the sample (int16)
    SKE (skin) events:
        0001 0000 TsRR Rbbb RRcR Rttt tttt tttp
        polarity(p) - bool
        taxel(t) - uint16
        cross_base(c) - bool - for events we can ignore this
        body_part(b) - uint8
        side(s) - uint8 (could be bool)
        type(T) - uint8 (could be bool) - 0 for events, 1 for samples
        reserved (R)
    EAR (Cochlear / audio) events:
        0000 0sss 0css snnn nnnn rrmo ffff fffp
        polarity          (p: 0 -> positive; 1 -> negative)
        frequency channel (f: from 0 to 31)
        olive model       (o: 0 -> MSO; 1 -> LSO)
        auditory model    (m: 0 -> cochlea; 1 -> superior olivary complex)
        reserved          (r: fixed to 0)
        ITD neurons id    (n: from 0 to 15)
        sensor ID         (s: fixed to b'100')
        channel           (c: 0 -> left; 1 -> right)
    """
    dvsBool = np.logical_not(data[:, 1] & 0xFF800000)
    apsBool = np.array(data[:, 1] & 0x00800000, dtype=bool)
    skinBool = np.array(data[:, 1] & 0x01000000, dtype=bool)
    imuBool = np.array(data[:, 1] & 0x02000000, dtype=bool)
    earBool = np.array(data[:, 1] & 0x04000000, dtype=bool)
    if kwargs.get('codec', '24bit') == '20bit':
        dvsBool[:] = True
        apsBool[:] = False
        skinBool[:] = False
        imuBool[:] = False
        earBool[:] = False

    outDict = {
        'dvs': None,
        'aps': None,
        'skinSamples': None,
        'skinEvents': None,
        'imuSamples': None,
        'ear': None,
    }
    if np.any(dvsBool):
        dataDvs = data[dvsBool, :]
        ts = dataDvs[:, 0]
        if np.isscalar(ts):
            ts = np.ones((1), dtype=np.uint32) * ts
        pol = ~np.array(dataDvs[:, 1] & 0x01, dtype=bool)  # We want True=ON=brighter, False=OFF=darker, so we negate
        dataDvs[:, 1] >>= 1
        if kwargs.get('codec', '24bit') == '20bit':
            # If any non zero value shows up in bits 18-19 then we are sure that the new 24 bit codec is being used
            if np.array(dataDvs[:, 1] & 0x00C00000, dtype=bool).any():
                raise ValueError("Data codec not consistent or data check not valid")
            x = np.uint16(dataDvs[:, 1] & 0x1FF)
            dataDvs[:, 1] >>= 9
            y = np.uint16(dataDvs[:, 1] & 0xFF)
            dataDvs[:, 1] >>= 10
        else:  # 24bit - default
            x = np.uint16(dataDvs[:, 1] & 0x3FF)
            dataDvs[:, 1] >>= 11
            y = np.uint16(dataDvs[:, 1] & 0x3FF)
            dataDvs[:, 1] >>= 10
        ch = np.uint8(dataDvs[:, 1] & 0x01)
        outDict['dvs'] = {
            'ts': ts,
            'ch': ch,
            'x': x,
            'y': y,
            'pol': pol}

    if np.any(imuBool):
        dataImu = data[imuBool, :]
        ts = np.uint32(dataImu[:, 0])
        if np.isscalar(ts):
            ts = np.ones((1), dtype=np.uint32) * ts
        value = int(dataImu[:, 1] & 0xFFFF)
        dataImu[:, 1] >>= 16
        sensor = np.uint8(dataImu[:, 1] & 0x0F)
        dataImu[:, 1] >>= 6
        ch = np.uint8(dataImu[:, 1] & 0x01)
        outDict['imuSamples'] = {
            'ts': ts,
            'ch': ch,
            'sensor': sensor,
            'value': value}

    if np.any(skinBool):
        dataSkin = data[skinBool, :]
        ts = np.uint32(dataSkin[:, 0])
        if np.isscalar(ts):
            ts = np.ones((1), dtype=np.uint32) * ts
        polarity = np.array(dataSkin[:, 1] & 0x01, dtype=bool)
        dataSkin[:, 1] >>= 1
        taxel = np.uint16(dataSkin[:, 1] & 0x3FF)
        dataSkin[:, 1] >>= 12
        #        crossBase = np.uint8(dataSkin[:, 1] & 0x01)
        dataSkin[:, 1] >>= 3
        bodyPart = np.uint8(dataSkin[:, 1] & 0x07)
        dataSkin[:, 1] >>= 6
        side = np.uint8(dataSkin[:, 1] & 0x01)
        #        dataSkin[:, 1] >>= 1
        #        eventType = np.uint8(dataSkin[:, 1] & 0x01)
        outDict['skinEvents'] = {
            'ts': ts,
            'side': side,
            'bodyPart': bodyPart,
            'taxel': taxel,
            'pol': polarity}

    if np.any(earBool):
        dataEar = data[earBool, :]
        ts = np.uint32(dataEar[:, 0])
        if np.isscalar(ts):
            ts = np.ones((1), dtype=np.uint32) * ts
        polarity = np.array(dataEar[:, 1] & 0x01, dtype=bool)
        dataEar[:, 1] >>= 1
        frequencyChannel = np.uint8(dataEar[:, 1] & 0x7F)
        dataEar[:, 1] >>= 7
        xsoType = np.uint8(dataEar[:, 1] & 0x01)
        dataEar[:, 1] >>= 1
        auditoryModel = np.uint8(dataEar[:, 1] & 0x01)
        dataEar[:, 1] >>= 3
        itdNeuronIds = np.uint8(dataEar[:, 1] & 0x7F)
        dataEar[:, 1] >>= 10
        channel = np.uint8(dataEar[:, 1] & 0x01)
        outDict['ear'] = {
            'ts': ts,
            'ch': channel,
            'itdNeuronIds': itdNeuronIds,
            'auditoryModel': auditoryModel,
            'xsoType': xsoType,
            'freq': frequencyChannel,
            'pol': polarity}
    return outDict


def getOrInsertDefault(inDict, arg, default):
    """
    Get the value specified by `arg` from the dictionary `inDict`.

    If `inDict` doesn't contain `arg`, return `default`.
    Also insert `default` into `inDict`.
    """
    value = inDict.get(arg, default)
    if value == default:
        inDict[arg] = default
    return value


def samplesToImu(inDict, **kwargs):
    """
    Group IMU samples together into multi-dimensional samples.

    `inDict` is a dictionary containing `ts`, `sensor`, and `value`.
    Assume these are IMU samples, and compile them (up to) 10 at a time.

    `outDict` is a dictionary containing `ts`, `acc`, `angV`, `temp` and `mag`.

    The `sensor` key is defined as:
        0: - (negative) Accel Y     1: Accel X     2: Accel Z
        3: - (negative) Gyro Y      4: Gyro X      5: Gyro Z
        6: Temperature
        7: - (negative) Mag Y       8: Mag X       9: Mag Z

    For the IMU on STEFI, (which is: ICM-20948):
        - gyro full scale is +/-2000 degrees per second
        - accelerometer full scale is +/-2 g
        - temp - 333.87 but does it zero at 0K or at room temperature? (+21deg)
        - mag full scale is +/- 4900 uT
    """
    accConversionFactor = kwargs.get('accConversionFactor', (32768 / 2) / 9.80665)
    angVConversionFactor = kwargs.get('angVConversionFactor', (32768 / 250) * (180 / np.pi))
    tempConversionFactor = kwargs.get('tempConversionFactor', 333.87)
    tempConversionOffset = kwargs.get('tempConversionOffset', -273.15 - 21)
    magConversionFactor = kwargs.get('magConversionFactor', (32768 / 4900) * 1000000)
    # Assume that sensor always ramps up, and split when it wraps around
    # Otherwise, iterate through this sample by sample - slowest and most robust method
    # Possible to do this much faster, but only my assuming no gaps in data
    tsAll = inDict['ts']
    sensorAll = inDict['sensor'].astype(int)
    valueAll = inDict['value']
    wrapIds = np.where((sensorAll[1:] - sensorAll[:-1]) < 1)[0]
    numImu = len(wrapIds) + 1
    tsOut = np.zeros((numImu), dtype=float)
    acc = np.zeros((numImu, 3), dtype=int)
    angV = np.zeros((numImu, 3), dtype=int)
    temp = np.zeros((numImu, 1), dtype=int)
    mag = np.zeros((numImu, 3), dtype=int)
    imuPtr = -1
    sensorPrev = 100
    for ts, sensor, value in zip(tsAll, sensorAll, valueAll):
        if sensor <= sensorPrev:
            imuPtr += 1
            tsOut[imuPtr] = ts  # Just take the first ts for a group of samples
        sensorPrev = sensor
        if sensor == 0:
            acc[imuPtr, 1] = -value
        elif sensor == 1:
            acc[imuPtr, 0] = value
        elif sensor == 2:
            acc[imuPtr, 2] = value
        elif sensor == 3:
            angV[imuPtr, 1] = -value
        elif sensor == 4:
            angV[imuPtr, 0] = value
        elif sensor == 5:
            angV[imuPtr, 2] = value
        elif sensor == 6:
            temp[imuPtr, 0] = value
        elif sensor == 7:
            mag[imuPtr, 1] = -value
        elif sensor == 8:
            mag[imuPtr, 0] = value
        elif sensor == 9:
            mag[imuPtr, 2] = value
    acc = acc.astype(float) / accConversionFactor
    angV = angV.astype(float) / angVConversionFactor
    temp = temp.astype(float) / tempConversionFactor - tempConversionOffset
    mag = mag.astype(float) / magConversionFactor
    outDict = {
        'ts': tsOut,
        'acc': acc,
        'angV': angV,
        'temp': temp,
        'mag': mag,
    }
    return outDict


def groupSkinSamples(inDict):
    """
    Group skin samples together into multi-dimensional samples.

    Similarly to imu data, skin samples are delivered in batches with each of
    the 192 taxels sampled once in a repeating pattern, each with its own time-
    stamp. Group these together into multi-dimensional samples each with a sin-
    gle timestamp. We use the observation that groups start with taxel 213 to
    simplify this.

    TODO: Use a more robust method of identifying groups.

    Note: We reject any data that doesn't form a complete group!

    The output dictionary includes a key `taxelOrder` so that the ordered sam-
    ples can later be mapped to their respective addresses.
    """
    wrapIds = np.where((inDict['taxel'] == 213))[0]
    pressure = inDict['value'][wrapIds[0]:wrapIds[-1]].reshape(-1, 192)
    taxelOrder = inDict['taxel'][wrapIds[0]:wrapIds[1]]
    ts = inDict['ts'][wrapIds[0]:wrapIds[-1]:192]
    return {'ts': ts,
            'pressure': pressure,
            'taxelOrder': taxelOrder}


def appendBatch(mainDict, batch):
    """
    Append `batch` of events to the dictionary `mainDict`.

    For a batch of events for a given `dataType`, and the container of all
    events for that `dataType` so far, add the batch to the container.

    The first batch for each data type gets used as the model to which all
    further batches get added.

    A container is a dict of numpy arrays. As these arrays need to be expanded
    to fit the new data, they are doubled in size. This limits to O(log(n)) the
    number of times we need to resize the arrays and it limits to nx4 the amount
    of memory used. An alternative would be to build lists and then convert to
    numpy at the end.
    """
    if batch is None:
        return mainDict
    numEventsBatch = len(batch['ts'])
    if mainDict is None:
        batch['numEvents'] = numEventsBatch
        return batch
    # Check if the main array has enough free space for this batch
    sizeOfMain = len(mainDict['ts'])
    numEventsMain = mainDict['numEvents']
    while numEventsBatch + numEventsMain > sizeOfMain:
        for fieldName in mainDict.keys():
            if fieldName != 'numEvents':
                # Double the size of the array
                mainDict[fieldName] = np.append(mainDict[fieldName], np.empty_like(mainDict[fieldName]), axis=0)
        sizeOfMain *= 2
    # Add the batch into the main array
    for fieldName in batch.keys():
        if mainDict[fieldName].ndim == 2:
            mainDict[fieldName][numEventsMain:numEventsMain + numEventsBatch, :] = \
                batch[fieldName]
        else:
            mainDict[fieldName][numEventsMain:numEventsMain + numEventsBatch] = \
                batch[fieldName]
    mainDict['numEvents'] = numEventsMain + numEventsBatch
    return mainDict


def cropArraysToNumEvents(inDict):
    """Crop each array of `inDict` up to the number of events in `inDict`."""
    if inDict is None:
        return None
    numEvents = inDict.pop('numEvents')
    for fieldName in inDict.keys():
        inDict[fieldName] = inDict[fieldName][:numEvents]
    return inDict


def globalPostProcessing(inDict, **kwargs):
    """
    Carry out global post-processing steps on the imported dictionary `inDict`.

    Main steps:
        1) Zero timestamps.
        2) Split by channel - The iit yarp format assumes that the channel bit
           corresponds to `left` and `right` sensors, so it's handled explici-
           tly here.
        3) Construct output dictionary.
    """
    # zero time by default (keeping the offset)
    if getOrInsertDefault(kwargs, 'zeroTimestamps', True):
        zeroTimestampsForAChannel(inDict)
    # Split into left and right channels
    # Allowing previously used 'split_stereo' for backwards compatibility
    splitByChannel = kwargs.get('splitChannel', kwargs.get('split_stereo', True))
    if splitByChannel:
        dataTypes = list(inDict.keys())
        channels = {}
        for chIdx, chName in enumerate(['left', 'right']):
            chDict = {}
            for dataType in dataTypes:
                dataForChannel = None
                if 'ch' in inDict[dataType]:
                    dataForChannel = selectByLabel(inDict[dataType], 'ch', chIdx)
                elif 'side' in inDict[dataType]:
                    dataForChannel = selectByLabel(inDict[dataType], 'side', chIdx)
                if dataForChannel is not None:
                    chDict[dataType] = dataForChannel
            if any(chDict):
                channels[chName] = chDict
        for dataType in dataTypes:
            if 'ch' in inDict[dataType] or 'side' in inDict[dataType]:
                del inDict[dataType]
        if any(inDict):
            channels['general'] = inDict
    else:
        channels = {'all': inDict}
    # Construct the output dict
    outDict = {
        'info': kwargs,
        'data': channels
    }
    outDict['info']['fileFormat'] = 'iityarp'
    return outDict


def importPostProcessing(inDict, **kwargs):
    """
    Carry out post-processing steps on the imported dictionary `inDict`.

    These are common steps for data imported both from `data.log` and from bin-
    ary file.

    Main steps:
        1) Eliminate dict keys which have no data
        2) Unwrap timestamps and convert to seconds
        3) IMU data special handling - convert samples to IMU (because up to 10
           consecutive samples define single imu read-out)
        4) Skin data special handling - ...
        5) Call to global post processing for splitting by channel, timestamp
           zeroing and formation as a top-level container.
    """
    # List out the datatypes here because the for loop changes the dict
    dataTypes = list(inDict.keys())
    for dataType in dataTypes:
        # Eliminate dataTypes which didn't have any events
        if inDict[dataType] is None or not dataType == 'dvs':  # TODO find a better way to handle not dvs data
            del inDict[dataType]
        else:
            # Iron out any time-wraps which occurred and convert to seconds
            clock_period_ns = kwargs.get('clock')
            if clock_period_ns is None:
                clock_period_s = 1 / 1e6
            else:
                clock_period_s = clock_period_ns / 1e9

            inDict[dataType]['ts'] = unwrapTimestamps(inDict[dataType]['ts'],
                                                      **kwargs) * clock_period_s
            # Special handling for imu data
            if dataType == 'imuSamples':
                if kwargs.get('convertSamplesToImu', True):
                    inDict['imu'] = samplesToImu(inDict.pop('imuSamples'), **kwargs)
                else:
                    inDict['sample'] = inDict.pop('imuSamples')
                    # TODO: The name "sample" is too vague
            # Special handling for skin data
            if dataType == 'skinSamples':
                inDict['skinSamples'] = groupSkinSamples(inDict['skinSamples'])
    return globalPostProcessing(inDict, **kwargs)


def importIitYarpBinaryDataLog(**kwargs):
    """Import data in IIT Yarp format from a binary file."""
    importFromByte = kwargs.get('importFromByte', 0)
    importMaxBytes = kwargs.get('importMaxBytes')
    if importMaxBytes is not None:
        importMaxBytes -= np.mod(importMaxBytes, 8)  # Events are 8 bytes long
    with open(kwargs['filePathOrName'], 'rb') as file:
        file.seek(importFromByte)
        if importMaxBytes is not None:
            events = file.read(importMaxBytes)
            # TODO: check if importMaxBytes is "feasible" (not beyond the file size)?
        else:
            events = file.read()
        events = np.frombuffer(events, np.uint32)
        events = events.reshape((-1, 2))
        outDict = decodeEvents(events)
        del events
        if file.read(1) is None:
            kwargs['importedToByte'] = 'EOF'
        else:
            kwargs['importedToByte'] = file.tell() - 2
            # - 2 compensates for the small read-ahead just performed to test EOF
    return importPostProcessing(outDict, **kwargs)


def importIitRawSkinSamples(**kwargs):
    """
    Import raw skin samples expressed in IIT Yarp format.

    Some files from the 'Old MTB' (As of 2020_12 mounted on iCub) where skin is
    dumped it creates a quite different bottle format:
        bottleNumber(int - actually alway -1) ts(float already in seconds) sample1(float) sample2(float) ...
    """
    importFromByte = kwargs.get('importFromByte', 0)
    importToByte = kwargs.get('importMaxBytes')
    if importToByte is not None:
        importToByte += importFromByte
    with open(kwargs['filePathOrName'], 'r') as file:
        file.seek(importFromByte)
        importedToByte = 'EOF'  # default if the following loop exits normally
        line = file.readline()
        currentPointer = 0
        skinSamples = None
        while line:
            if importToByte is not None:
                if file.tell() > importToByte:
                    importedToByte = currentPointer - 1
                    break
                else:
                    currentPointer = file.tell()
            try:
                numbers = line.split(' ')
                numSamples = len(numbers) - 2
                # TODO: Later, we might generalise the number of channels.
                assert numSamples == 192
                ts = float(numbers[1])
                samples = np.array(numbers[2:], dtype=float)[np.newaxis, :]
                newSample = {'ts': np.ones((1), dtype=float) * ts,
                             'pressure': samples}
                skinSamples = appendBatch(skinSamples, newSample)
            except ValueError:  # sometimes finding malformed packets at the end of files - ignoring
                print('Failed to interpret a line')
                line = file.readline()
                continue
            line = file.readline()
    skinSamples['ts'] = unwrapTimestamps(skinSamples['ts'])  # will this ork on float values?
    outDict = {'skinSamples': cropArraysToNumEvents(skinSamples)}
    kwargs['importedToByte'] = importedToByte
    return globalPostProcessing(outDict, **kwargs)

def unquoting(match):
    matchedString = (match.string[match.span()[0]:match.span()[1]])
    even = len(matchedString) % 2 == 0
    if matchedString[-2:] == b'\\0':
        return b'\\' * ((len(matchedString) - 1) // 2) + b'\0' if even\
            else b'\\' * ((len(matchedString) - 1) // 2) + b'0'
    if matchedString[-2:] == b'\\n':
        return b'\\' * ((len(matchedString) - 1) // 2) + b'\n' if even\
            else b'\\' * ((len(matchedString) - 1) // 2) + b'n'
    if matchedString[-2:] == b'\\r':
        return b'\\' * ((len(matchedString) - 1) // 2) + b'\r' if even\
            else b'\\' * ((len(matchedString) - 1) // 2) + b'r'
    if matchedString[-2:] == b'\\"':
        return b'\"'
    if matchedString[-2:] == b'\\\\':
        if match.string[match.span()[1]:match.span()[1]+1] in [b'0', b'n', b'r', b'\0', b'\n', b'\r']:
            return matchedString
        return b'\\' * ((len(matchedString)) // 2)


def fromStringNested(in_string):
   return re.sub(b'\\\{2,}|\\\\\"', unquoting, re.sub(b'\\\\+[nr0]', unquoting, in_string))



def importIitYarpDataLog(**kwargs):
    """Import data in IIT Yarp format from a `data.log` file."""
    # Check if format suggests data from vicon dumper
    # Create dicts for each possible datatype
    dvs = None
    dvsLbl = None
    dvsFlow = None
    imuSamples = None  # Imu Sample is an intermediate datatype - later it gets converted to IMU etc
    skinEvents = None
    skinSamples = None
    ear = None
    aps = None
    importFromByte = kwargs.get('importFromByte', 0)
    importToByte = kwargs.get('importMaxBytes')
    importedToByte = 'EOF'  # default if the following loop exits normally

    try:
        with open(kwargs['filePathOrName'], 'r') as inFile:
            content = inFile.readline()  # Look at first line of file
            patternForVicon = re.compile('(\d+) (\d+\.\d+) \((.*)\)')
            if patternForVicon.findall(content):
                print('Yarp vicon dumper pattern found - passing this file to importVicon function')
                return importIitVicon(**kwargs)
            patternForRawSkinSamples = re.compile('(\d+) (\d+\.\d+) (\d+\.\d+)')
            if patternForRawSkinSamples.findall(content):
                print('Pattern found seems to be a raw dump of analogue skin samples')
                return importIitRawSkinSamples(**kwargs)

        pattern = re.compile('(\d+) (\d+\.\d+) ([A-Z]+) \((.*)\)')
        if importToByte is not None:
            importToByte += importFromByte
        with open(kwargs['filePathOrName'], 'r') as file:
            file.seek(importFromByte)
            line = file.readline()
            currentPointer = 0
            while line:
                if importToByte is not None:
                    if file.tell() > importToByte:
                        importedToByte = currentPointer - 1
                        break
                    else:
                        currentPointer = file.tell()
                found = pattern.match(line)
                # The following values would be useful for indexing the input file:
                # bottlenumber = np.uint32(found[1])
                # timestamp = float(found[2])
                bottleType = found[3]
                if bottleType not in ['AE', 'IMUS', 'LAE', 'FLOW', 'EAR', 'SKS', 'SKE']:
                    print('Unknown bottle type: ' + bottleType)
                    line = file.readline()
                    continue
                try:
                    events = np.array(found[4].split(' '), dtype=np.uint32)
                    '''
                    The following block handles unusual bottle format for labelled
                    events and flow events, which all end up being treated as
                    'dvs' type events. In those cases there are extra words in
                    the bottle for each event.
                    '''
                    if bottleType == 'LAE':
                        numEventsInBatch = int(len(events) / 3)
                        events = events.reshape(numEventsInBatch, 3)
                        outDict = decodeEvents(events[:, :2], **kwargs)
                        dvsBatch = outDict['dvs']
                        dvsBatch['lbl'] = events[:, 2]
                        dvsLbl = appendBatch(dvsLbl, dvsBatch)
                    elif bottleType == 'FLOW':
                        numEventsInBatch = int(len(events) / 4)
                        events = events.reshape(numEventsInBatch, 4)
                        outDict = decodeEvents(events[:, :2], **kwargs)
                        dvsBatch = outDict['dvs']
                        dvsBatch['vx'] = events[:, 2].view(dtype=float)
                        dvsBatch['vy'] = events[:, 3].view(dtype=float)
                        dvsFlow = appendBatch(dvsFlow, dvsBatch)
                    elif bottleType == 'SKS':
                        ''' Skin samples are encoded with 4 words:
                            [ts, address, ts, analogue_sample(least-sig. 16 bits)]
                            The address is encoded the same as for skin events
                            we ignore the second timestamp
                        '''
                        numEventsInBatch = int(len(events) / 4)
                        events = events.reshape(numEventsInBatch, 4)
                        outDict = decodeEvents(events, **kwargs)
                        skinSampleBatch = outDict.pop(
                            'skinEvents')  # pop these so they don't get treated as skin events, below
                        skinSampleBatch['value'] = np.mod(events[:, 3], 2 ** 16)
                        skinSamples = appendBatch(skinSamples, skinSampleBatch)
                    else:  # bottleType in ['AE', 'IMUS', 'SKE', 'EAR', 'APS']
                        numEventsInBatch = int(len(events) / 2)
                        events = events.reshape(numEventsInBatch, 2)
                        outDict = decodeEvents(events[:, :2], **kwargs)
                        dvsBatch = outDict['dvs']
                        dvs = appendBatch(dvs, dvsBatch)
                        imuSamples = appendBatch(imuSamples, outDict['imuSamples'])
                        ear = appendBatch(ear, outDict['ear'])
                        skinEvents = appendBatch(skinEvents, outDict['skinEvents'])
                        aps = appendBatch(aps, outDict['aps'])
                except ValueError:  # sometimes finding malformed packets at the end of files - ignoring
                    line = file.readline()
                    continue
                line = file.readline()
    except UnicodeDecodeError:
        with open(kwargs['filePathOrName'], 'rb') as file:
            content = file.readlines()
        eventsToDecode = []
        timestamps = []
        check_if_with_ts = False
        for c in tqdm(content):
            firstQuoteIdx = c.find(b'\"')
            lastQuoteIdx = c[::-1].find(b'\"')
            bottleNum, ts, bottleType, _ = c[:firstQuoteIdx - 1].decode().split(' ')
            data = c[firstQuoteIdx + 1:-(lastQuoteIdx + 1)]
            bitStrings = np.frombuffer(fromStringNested(data), np.uint32)
            if not check_if_with_ts:
                with_ts = np.all(sorted(bitStrings[::2]) == bitStrings[::2])
                check_if_with_ts = True
            eventsToDecode.append(bitStrings)
            timestamps += [float(ts) / 0.000001]*len(bitStrings)
        if with_ts:
            outDict = decodeEvents(np.reshape(np.concatenate(eventsToDecode), (-1, 2)))
        else:
            outDict = decodeEvents(np.vstack((timestamps, np.concatenate(eventsToDecode))).swapaxes(0, 1).astype(int))
        del content
        return importPostProcessing(outDict, **kwargs)


    # If importedToByte is not defined then we reached the end of the file
    # Crop arrays to number of events
    outDict = {
        'dvs': cropArraysToNumEvents(dvs),
        'dvsLbl': cropArraysToNumEvents(dvsLbl),
        'flow': cropArraysToNumEvents(dvsFlow),  # TODO: 'flow' is a poor name, considering we also have dense flow maps
        'imuSamples': cropArraysToNumEvents(imuSamples),
        # Imu Sample is an intermediate datatype - later it gets converted to IMU etc
        'skinEvents': cropArraysToNumEvents(skinEvents),
        'skinSamples': cropArraysToNumEvents(skinSamples),
        'ear': cropArraysToNumEvents(ear),
        'aps': cropArraysToNumEvents(aps),
    }
    kwargs['importedToByte'] = importedToByte
    return importPostProcessing(outDict, **kwargs)


def importIitYarpInfoLog(**kwargs):
    """
    Import `info.log` from data collected in IIT Yarp format.

    From the `info.log` we want the time that the channel connected; we assume
    this is the first decimal found.
    """
    # Read file
    filePathOrName = kwargs.get('filePathOrName')
    print('Examining info.log: ' + filePathOrName)
    with open(filePathOrName, 'r') as inFile:
        content = inFile.read()
    patternForInfoLog = re.compile('(\d+\.\d+)')
    found = patternForInfoLog.search(content)
    if found:
        tsOffset = float(found[0])
        return tsOffset


def importIitYarpRecursive(**kwargs):
    """
    Import data in IIT Yarp format in a recursive manner.

    This function works in the following way for efficiency when importing very
    large files:
    - Import events bottle by bottle
    - Create arrays to hold the first 1024 values then extend the array size
      exponentially while importing, finally cropping these arrays.

    Note: `kwargs` is augmented where necessary and becomes the "info" dict of
    the output.
    """
    path = getOrInsertDefault(kwargs, 'filePathOrName', '.')
    print('importIitYarp trying path: ' + path)
    if not os.path.exists(path):
        raise FileNotFoundError("path not found.")
    if not os.path.isdir(path):
        raise FileNotFoundError("path is not a directory.")
    files = sorted(os.listdir(path))
    importedDicts = []
    tsOffset = None
    boundingBoxes = None
    skeleton = None
    for file in files:
        filePathOrName = os.path.join(path, file)
        kwargs['filePathOrName'] = filePathOrName
        if os.path.isdir(filePathOrName):
            importedDicts = importedDicts + importIitYarpRecursive(**kwargs)
        if file == 'binaryevents.log':
            importedDicts.append(importIitYarpBinaryDataLog(**kwargs))
        if file == 'data.log':
            importedDicts.append(importIitYarpDataLog(**kwargs))
        if file == 'info.log':
            tsOffset = importIitYarpInfoLog(**kwargs)
        if file == 'ground_truth.csv':
            boundingBoxes = importBoundingBoxes(**kwargs)
        if file == 'skeleton.json':
            skeleton = importSkeleton(**kwargs)
    if len(importedDicts) == 0:
        print('    "data.log" file not found')
        return importedDicts
    if tsOffset is not None:
        importedDicts[-1]['info']['tsOffsetFromInfo'] = tsOffset
    if boundingBoxes is not None:
        addGroundTruth(boundingBoxes, importedDicts, 'boundingBoxes')
    if skeleton is not None:
        addGroundTruth(skeleton, importedDicts, 'skeleton')
    return importedDicts


def addGroundTruth(groundTruth, importedDicts, name):
    if len(importedDicts) == 1:
        keys = list(importedDicts[-1]['data'])
        if len(keys) == 1:
            importedDicts[-1]['data'][keys[0]][name] = groundTruth
            try:
                importedDicts[-1]['data'][keys[0]][name]['tsOffset'] = -importedDicts[-1]['info']['tsOffsetFromInfo'] # TODO check if this is always valid
            except KeyError:
                pass
        else:
            # TODO If more than one channel is present we don't know which one to assign the ground truth
            print(f'Found channels {keys}. Don\'t know which one to assign ground truth. Skipping.')


def importIitYarp(**kwargs):
    """Import data in IIT Yarp format."""
    importedDicts = importIitYarpRecursive(**kwargs)
    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
        # Optional: start the timestamps at zero for the first event
        # This is done collectively for all the concurrent imports
        rezeroTimestampsForImportedDicts(importedDicts)
    if len(importedDicts) == 1:
        importedDicts = importedDicts[0]
    return importedDicts



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

In general get_frame takes two args: time, and time_window.
In general one could think about representing data in an interpolated way or not.
For example, with poses, imu etc, one could interpolate between samples, 
or one could simply choose the sample which is nearest in time.
Likewise for frames. 
The time_window parameter says how much data to take around the sample point for event-type data.
It might be possible to develop visualisations for other types of data that make use of the concept of a time window. 

colorfmt is a choice between luminance and rgb. 
If luminance, then the frame returned should have dim 2 = 3.
Nothing stops the calling function from applying a color mask to an output in luminance format.      
"""

import numpy as np

# Local imports

# Importing child classes lets this one module be referenced for all imports



try:
    from .visualisers.visualiserDvs import VisualiserDvs
    from .visualisers.visualiserFrame import VisualiserFrame
    from .visualisers.visualiserPoint3 import VisualiserPoint3
    from .visualisers.visualiserPose6q import VisualiserPose6q
    from .visualisers.visualiserBoundingBoxes import VisualiserBoundingBoxes
    from .visualisers.visualiserOpticFlow import VisualiserOpticFlow
    from .visualisers.visualiserImu import VisualiserImu
    from .visualisers.visualiserSkeleton import VisualiserSkeleton
except ImportError:
    from bimvee.visualisers.visualiserDvs import VisualiserDvs
    from bimvee.visualisers.visualiserFrame import VisualiserFrame
    from bimvee.visualisers.visualiserPoint3 import VisualiserPoint3
    from bimvee.visualisers.visualiserPose6q import VisualiserPose6q
    from bimvee.visualisers.visualiserBoundingBoxes import VisualiserBoundingBoxes
    from bimvee.visualisers.visualiserOpticFlow import VisualiserOpticFlow
    from bimvee.visualisers.visualiserImu import VisualiserImu
    from bimvee.visualisers.visualiserSkeleton import VisualiserSkeleton




# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

VisualiserFrame visualises frame data, using the frame in the dict nearest
to the current time.

There is an optional mode: you can add a 'tsEnd' field to the dict, an np array
of length correspinding to 'ts'. Then, the selected frame will be the last
one prior to (or exactly equal to) the current time, and will only be shown
until 'tsEnd'. Otherwise, timeWindow dictates whether the selected (nearest)
frame is shown or not.
"""

import numpy as np
import math

# Local imports
from .visualiserBase import Visualiser

# A function intended to find the nearest timestamp
# adapted from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
def findNearest(array, value):
    idx = np.searchsorted(array, value) # side="left" param is the default
    if idx > 0 and ( \
            idx == len(array) or \
            math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
        return idx-1
    else:
        return idx


class VisualiserFrame(Visualiser):

    data_type = 'frame'

    def set_data(self, data):
        self.__data = {}
        self.__data.update(data)

        if self.__data['frames'][0].dtype != np.uint8:
            # Convert to uint8, converting to fullscale accross the whole dataset
            minValue = min([frame.min() for frame in self.__data['frames']])
            maxValue = max([frame.max() for frame in self.__data['frames']])
            self.__data['frames'] = [((frame-minValue)/(maxValue-minValue)*255).astype(np.uint8) for frame in data['frames']] #TODO: assuming that it starts scaled in 0-1 - could have more general approach?

    def get_colorfmt(self):
        try:
            if len(self.__data['frames'][0].shape) == 3:
                return 'rgb'
            else:
                return 'luminance'
        except: # TODO None type error?
            return 'luminance'
            
    def get_default_image(self):
        x, y = self.get_dims()
        # Return an x,y,3 by default i.e. rgb, for safety, since in the absence of data we may not know how the texture's colorfmt is set
        return np.ones((x, y, 3), dtype=np.uint8) * 128 # TODO: Hardcoded midway (grey) value

    # TODO: There can be methods which better choose the best frame, or which create a visualisation which
    # respects the time_window parameter 
    def get_frame(self, time, timeWindow, **kwargs):
        data = self.__data
        if 'tsEnd' in data:
            # Optional mode in which frames are only displayed 
            # between corresponding ts and tsEnd
            frameIdx = np.searchsorted(data['ts'], time, side='right') - 1
            if frameIdx < 0:
                image = self.get_default_image()
            elif time > data['tsEnd'][frameIdx]:
                image = self.get_default_image()
            else:                
                image = data['frames'][frameIdx]
        elif time < data['ts'][0] - timeWindow / 2 or time > data['ts'][-1] + timeWindow / 2:
            # Gone off the end of the frame data
            image = self.get_default_image()
        else:
            frameIdx = findNearest(data['ts'], time)
            image = data['frames'][frameIdx]
        # Allow for arbitrary post-production on image with a callback
        # TODO: as this is boilerplate, it could be pushed into pie syntax ...
        if kwargs.get('callback', None) is not None:
            kwargs['image'] = image
            image = kwargs['callback'](**kwargs)
        return image
    
    def get_dims(self):
        try:
            data = self.__data
        except AttributeError: # data hasn't been set yet
            return 1, 1
        x = data['dimX'] if 'dimX' in data else data['frames'][0].shape[1]
        y = data['dimY'] if 'dimY' in data else data['frames'][0].shape[0]
        return x, y




# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Takes a dict as imported by importAe, and for each channel, 
tries to run the appropriate general visualisation function
"""

# local imports
from .plotDvsContrast import plotDvsContrast
from .plotEventRate import plotEventRate
from .plotFrame import plotFrame
from .plotPose import plotPose
from .plotImu import plotImu
from .plotFlow import plotFlow
from .container import Container

def plot(inDict, **kwargs):
    if isinstance(inDict, Container):
        inDict = inDict.container
    
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plot(inDictInst, **kwargs)
        return
    if isinstance(inDict, dict):
        if 'info' in inDict: # Special handling for a file-level container
            kwargs['title'] = inDict['info'].get('filePathOrName', '')
            plot(inDict['data'], **kwargs)
        else:
            title = kwargs.get('title', '')
            for keyName in inDict.keys():
                if isinstance(inDict[keyName], dict):
                    if keyName == 'dvs':
                        plotDvsContrast(inDict[keyName], **kwargs)
                        plotEventRate(inDict[keyName], **kwargs)
                    elif keyName == 'frame':
                        plotFrame(inDict[keyName], **kwargs)
                    elif keyName == 'imu':
                        plotImu(inDict[keyName], **kwargs)
                    elif keyName == 'pose6q':
                        plotPose(inDict[keyName], **kwargs)
                    elif keyName == 'point3' in channel:
                        plotPose(inDict[keyName], **kwargs)
                    elif keyName == 'flow' in channel:
                        plotFlow(inDict[keyName], **kwargs)
                    else:
                        kwargs['title'] = (title + '-' + keyName).lstrip('-')
                        plot(inDict[keyName], **kwargs)
